#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt

import os
from cortix import Cortix

def test_droplet():
    '''
    Run the droplet example
    '''
    pwd = os.path.dirname(__file__)
    full_path_config_file = os.path.join(pwd, 'input/cortix-config-droplet.xml')
    cortix1 = Cortix('cortix-droplet', full_path_config_file)
    cortix1.run_simulations(task_name="droplet-fall")
    os.system('rm -f *.png')
    os.system('rm -f *.pickle')

if __name__ == "__main__":
    test_droplet()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
Cortix: a program for system-level modules coupling, execution, and analysis.

Cortix is a library and it is used by means of a driver. This file is a simple example
of a driver. Many Cortix objects can be ran simultaneously; a single object
may be sufficient since many simulation/tasks can be ran via one object.

As Cortix evolves additional complexity may be added to this driver and/or other
driver examples can be created.
"""
#*********************************************************************************
import os
from cortix import Cortix 
#*********************************************************************************

def main():

 pwd                   = os.path.dirname(__file__)
 full_path_config_file = os.path.join(pwd, '../input/cortix-config.xml')

 # NB: if another instantiation of Cortix occurs, the cortix wrk directory specified
 #     in the cortix configuration file must be different, else the logging facility 
 #     will have log file collision.

 cortix1 = Cortix( 'cortix-dev1', full_path_config_file ) # see cortix-config.xml
#
# usage: uncomment one at a time.
 cortix1.run_simulations( task_name='solo-pyplot' )
# cortix1.run_simulations( task_name='solo-fueldepot' )
# cortix1.run_simulations( task_name='fueldepot-chopper' )
#---------------------- end def main():-------------------------------------------

#*********************************************************************************
# Usage: -> python cortix-main.py or ./cortix-main.py
if __name__ == "__main__":
 main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import os
from cortix import Cortix
#*********************************************************************************

def run(task_name=None):
    '''
    Run the Cortix Droplet example in a Jupyter Notebook for the following tasks:
     1. solo-droplet
     2. solo-pyplot
     3. droplet-fall

    In a Jupyter Notebook cell enter:
   `    import cortix.examples.main.main_droplet as droplet
        droplet.run(`task_name`)
    '''
    pwd = os.path.dirname(__file__)
    full_path_config_file = os.path.join(pwd, '../input/cortix-config-droplet.xml')
    cortix = Cortix('cortix-droplet', full_path_config_file)

    assert task_name == 'droplet-fall' or task_name == 'solo-droplet' or\
            task_name == 'solo-pyplot','FATAL: task name %r invalid.'%task_name

    cortix.run_simulations(task_name=task_name)

    return

#*********************************************************************************
if __name__ == "__main__":
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import os
from cortix import Cortix
#*********************************************************************************

def run():
    '''
    Run the Cortix Wind example. If Cortix and its dependencies are installed,
    this program should be executed at the command prompt inside the directory
    this program resides, namely, `$CORTIX/cortix/cortix/examples/console_run/`
    directory.
    '''

    pwd = os.path.dirname(__file__)
    full_path_config_file = os.path.join(pwd, '../input/cortix-config-wind.xml')

    cortix = Cortix('cortix-wind', full_path_config_file)

    cortix.run_simulations(task_name="solo-wind")

    return

#*********************************************************************************
if __name__ == "__main__":
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Network class for the Cortix project. A network defines the connectivity between
Cortix modules.

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import networkx as nx
from cortix.src.utils.xmltree import XMLTree
#*********************************************************************************

class Network:
    '''
    Cortix network class definition. Network class members:

    __config_node: XMLTree
        Configuration data in the form of an XML tree.

    __name:str
        Name of the network.

    __connectivity: list(dict)
        List of dictionaries of connectivity. Dictionary:
        {'fromModuleSlot': module_slot_name, 'fromPort': use_port_name, 'toModuleSlot': module_slot_name, 'toPort': provide_port_name}.

    __module_slot_names: list(str)
        List of names of module slots.

    __runtime_cortix_comm_file_name: dict
        Full path filename of the communication file for each module slot.
        {module_slot_name:full_path_comm_file_name, .:., .:., ...}.
        This is initially filled with a null filename at construction time. Later,
        Simulation.__setup_task() will fill in the information.

    __nx_graph:

    '''

#*********************************************************************************
# Construction 
#*********************************************************************************

    def __init__(self, net_config_node): 
        assert isinstance(net_config_node, XMLTree), '-> net_config_node is invalid.'

        self.__config_node = net_config_node

        assert self.__config_node.get_node_tag() == 'network'

        self.__name = self.__config_node.get_node_attribute('name')

        self.__connectivity = list(dict())
        self.__module_slot_names = list()

        # cortix communication file for modules
        self.__runtime_cortix_comm_file_name = dict()

        # network graph
        self.__nx_graph = nx.MultiDiGraph(name=self.__name)

        for child in self.__config_node.get_node_children():

            (element, tag, attributes, text) = child

            # a connect element must have no content; format: <connect />
            if tag == 'connect':
                assert text is None, 'non empty text, %r, in %r network: ' \
                    % (text, self.__name)

            tmp = dict()

            if tag == 'connect':

                for (key, value) in attributes:
                    assert key not in tmp.keys(), \
                        'repeated key in attribute of %r network' % self.__name
                    value = value.strip()
                    if key == 'fromModuleSlot':
                        value = value.replace(':', '_')
                    if key == 'toModuleSlot':
                        value = value.replace(':', '_')
                    tmp[key] = value

                self.__connectivity.append(tmp)

                for (key, val) in tmp.items():
                    if key == 'fromModuleSlot':
                        self.__runtime_cortix_comm_file_name[val] = \
                            'null-runtime_cortix_comm_file_name'
                    if key == 'toModuleSlot':
                        self.__runtime_cortix_comm_file_name[val] = \
                            'null-runtime_cortix_comm_file_name'
                vtx1 = tmp['fromModuleSlot']
                vtx2 = tmp['toModuleSlot']

                self.__nx_graph.add_edge(vtx1, vtx2,
                    fromPort=tmp['fromPort'], toPort=tmp['toPort'])

        self.__module_slot_names = [name for name in self.__runtime_cortix_comm_file_name.keys()]

        return

#*********************************************************************************
# Public member functions 
#*********************************************************************************

    def set_runtime_cortix_comm_file_name(self, module_slot_name, full_path_file_name):
        '''
        Sets the runtime cortix communications file to the one specified
        by full_path_file_name
        '''

        self.__runtime_cortix_comm_file_name[module_slot_name] = full_path_file_name

        return

    def get_runtime_cortix_comm_file_name(self, module_slot_name):
        '''
        Returns the cortix comm file that corresponds to module_slot_name. None if otherwise.
        '''

        if module_slot_name in self.__runtime_cortix_comm_file_name:
            return self.__runtime_cortix_comm_file_name[module_slot_name]
        else:
            return None

    def __get_name(self):
        '''
        `str`:Network name
        '''

        return self.__name

    name = property(__get_name, None, None, None)

    def __get_connectivity(self):
        '''
        `list(dict)`:List of the network connectivity
        '''

        return self.__connectivity

    connectivity = property(__get_connectivity, None, None, None)

    def __get_module_slot_names(self):
        '''
        `list(str)`:List of network slot names
        '''

        return self.__module_slot_names

    module_slot_names = property(__get_module_slot_names, None, None, None)

    def __get_nx_graph(self):
        '''
        `networkx.MultiDiGraph`:NXGraph corresponding to network
        '''

        return self.__nx_graph

    nx_graph = property(__get_nx_graph, None, None, None)

    def __str__(self):
        '''
        Network to string conversion
        '''

        s = 'Network data members:\n name=%s\n module slot names=%s\n connectivity=%s\n runtime comm file= %s'
        return s % (self.__name, self.__module_slot_names, self.__connectivity,
                    self.__runtime_cortix_comm_file_name)

    def __repr__(self):
        '''
        Network to string conversion
        '''

        s = 'Network data members:\n name=%s\n module slot names=%s\n connectivity=%s\n runtime comm file= %s'
        return s % (self.__name, self.__module_slot_names, self.__connectivity,
                    self.__runtime_cortix_comm_file_name)

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

#======================= end class Network: ======================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt

import importlib


def test_pandas():
    '''
    Test for the pandas dependency
    '''
    try:
        import pandas
        found = True
    except ImportError:
        found = False
    assert found == True


if __name__ == "__main__":
    test_pandas()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Cortix driver for guest modules.
Module developers must implement the public methods in this driver.
Ideally, this implementation should be minimal.
Developers should use this class to wrap their module (MyModule) implemented in a
file named my_module.py. This file will be placed inside the developer's module
directory which is pointed to in the Cortix config.xml file.
'''
#*************************************************************************
import os
import sys
import io
import time
import datetime
import logging

# uncomment
#from .my_module import MyModule
#*************************************************************************

class CortixDriverTemplate():
    '''
     Cortix driver for guest modules.
    '''

    def __init__(self,
                 slot_id,
                 input_full_path_file_name,
                 exec_full_path_file_name,
                 work_dir,
                 ports=list(),
                 cortix_start_time=0.0,
                 cortix_final_time=0.0,
                 cortix_time_unit=None,
                 cortix_time_step=0.0
                 ):

        # Sanity test
        assert isinstance(
            slot_id, int), '-> slot_id type %r is invalid.' % type(slot_id)
        assert isinstance(
            ports, list), '-> ports type %r is invalid.' % type(ports)
        assert len(ports) > 0
        assert isinstance(
            cortix_start_time, float), '-> time type %r is invalid.' % type(cortix_start_time)
        assert isinstance(
            cortix_final_time, float), '-> time type %r is invalid.' % type(cortix_final_time)

        assert isinstance(
            cortix_time_step, float), '-> time step type %r is invalid.' % type(cortix_time_step)

        assert isinstance(
            cortix_time_unit, str), '-> time unit type %r is invalid.' % type(cortix_time_unit)

        # Logging
        self.__log = logging.getLogger( 'launcher-mymodule' + str(slot_id) +
                                        '.cortixdriver')
        self.__log.info('initializing an object of CortixDriver()')

        # Guest library module: MyModule
        # uncomment
        # self.my_module = MyModule( slot_id, input_full_path_file_name, work_dir, ports,
        #                            cortix_start_time, cortix_final_time )

        self.__time_stamp = None  # temporary

        return
#---------------------- end def __init__():-------------------------------

    def call_ports(self, cortix_time=0.0):
        '''
        Call all ports at cortix_time
        '''

        self.__log_debug(cortix_time, 'call_ports')

        # uncomment
        # self.my_module.call_ports( cortix_time )

        self.__log_debug(cortix_time, 'call_ports')

        return
#---------------------- end def call_ports():-----------------------------

    def execute(self, cortix_time=0.0, timeStep=0.0):
        '''
        Evolve system from cortix_time to cortix_time + timeStep
        '''

        self.__log_debug(cortix_time, 'execute')

        # uncomment
        # self.my_module.execute( cortix_time, timeStep )

        self.__log_debug(cortix_time, 'execute')

        return
#---------------------- end def execute():--------------------------------

#*************************************************************************
# Private helper functions (internal use: __)

    def __log_debug(self, cortix_time=0.0, caller='null-function-name'):

        if self.__time_stamp is None:
            s = ''
            self.__log.debug(s)
            s = '=========================================================='
            self.__log.debug(s)
            s = 'CORTIX::DRIVER->***->DRIVER->***->DRIVER->***->DRIVER->***'
            self.__log.debug(s)
            s = '=========================================================='
            self.__log.debug(s)

            s = caller + '(' + str(round(cortix_time, 2)) + '[min]):'
            self.__log.debug(s)

            self.__time_stamp = time.time()

        else:

            end_time = time.time()

            s = caller + '(' + str(round(cortix_time, 2)) + '[min]): '
            m = 'CPU elapsed time (s): ' + \
                str(round(end_time - self.__time_stamp, 2))
            self.__log.debug(s + m)

            self.__time_stamp = None
            if caller == 'execute':
                s = ''
                self.__log.debug(s)

        return
#---------------------- end def __log_debug():----------------------------

#====================== end class CortixDriver: ==========================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
This file contains the class definition of `XMLTree`,
which aids in parsing the XML configuration files used
within the Cortix project.

Background
----------
An XML document consists of a collection of elements (nested or not). These elements
are containers of information and are the most important components of an XML
document. An element container is defined by a start tag and an end tag. A tag
starts with an opening angle bracket and finishes with a closing angle bracket as
follows
 + **element**: <ele_name> </ele_name>, *e.g*, <net_x> </net_x>.
Here `<net_x>` is the start tag and `</net_x>` is the end tag for the element named
`net_x`. An element with the same name can be used repeatedly in an XML document.
In `Cortix` we will use element names following the Python variable name snake
convention. It is also common to call the element name as the tag name. The start tag
of an element may have any number of atributes:
 + **attributes**: name='value', *e.g.*, <net_x mod_slot='wind:0'> </net_x>
these are name-value pairs; we will use the same name convention for element names
when creating attribute names. There may be many attributes in a start tag. Finally
everything in between tags of an element, is considered as the element content:
 + **content**: <tag> content </tag>, *e.g.*, <net_x mod_slot='wind:0'>189.67 MeV</net_x>

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
from xml.etree.ElementTree import ElementTree # the whole XML tree data structure
from xml.etree.ElementTree import Element     # an XML element or tree node
#*********************************************************************************

class XMLTree:
    '''
    This class is a wrapper around the XML parser ElementTree and Element. See
    import statement above. This XML parser is fast but the interface is not very
    user friendly; hence the motivation for this wrapper.
    The interface is designed to facilitate the use of XML data within Cortix and
    its modules. This class generates objects that hold an XML tree: ElementTree.
    Configuration of Cortix and some runtime files are the primary usage of XMLTree.
    The construction of an XMLTree object either uses a file with an XML content or
    an XML branch of an XML tree. This makes it useful throughout Cortix to inspect
    branches of a configuration XML tree to retrieve data.
    A node in a tree is the root of a branch. That is, the same thing as an XML
    element and all its direct sub-elements; described in the Background above.
    '''

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self, xml_tree_node=None, xml_tree_file=None):
        '''
        Overall description of the constructor
        '''

        # hold a given XML tree node
        if xml_tree_node is not None:
            assert isinstance(xml_tree_node, Element), '-> xml_tree_node invalid.'
            tag_name = xml_tree_node.tag
            assert isinstance(tag_name,str)
            assert len(tag_name) > 0

            self.__xml_tree_node = xml_tree_node

        # parse an XML tree and hold its root node
        if xml_tree_file is not None:
            assert isinstance(xml_tree_file, str), '-> configFileName not a str.'
            assert xml_tree_node is None, 'node and file not allowed together.'

            tree = ElementTree()
            tree.parse( xml_tree_file )

            self.__xml_tree_node = tree.getroot()

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def get_root_node(self):
        '''
        Returns the Element tree's root node.

        Parameters
        ----------
        empty:

        Returns
        -------
        self.__xml_tree_node: Element
        '''

        return self.__xml_tree_node

    def get_node_tag(self):
        '''
        Returns the tag name associated with the root node of the element tree. This is
        the element name or tag name, *e.g.*, `<elem_name> </elem_name>`.

        Parameters
        ----------
        empty:

        Returns
        -------
        tag_name: str
        '''

        tag_name = self.__xml_tree_node.tag
        assert isinstance(tag_name,str)

        return tag_name

    def get_node_attribute(self, attribute_name):
        '''
        Returns the value of the attribute associated with the root node of
        the element tree, *e.g.* <module type='native'></module>. Attribute
        name is `type`, value is 'native'.

        Parameters
        ----------
        attribute_name: str

        Returns
        -------
        attribute_value: str
        '''

        assert isinstance(attribute_name,str)
        attribute_value = self.__xml_tree_node.get(attribute_name.strip())

        assert attribute_value is not None
        assert isinstance(attribute_value,str)

        return attribute_value.strip()

    def get_node_content(self):
        '''
        Returns the content or text of the XML element.This is what is in between
        the start and end tags of the element.

        Parameters
        ----------
        Empty

        Returns
        -------
        content: str
        '''

        content = self.__xml_tree_node.text

        assert isinstance(content,str)

        return content

    def get_sub_node(self, tag):
        '''
        Returns the first subnode (branch) of the element tree specified by the
        parameter tag.

        Parameters
        ----------
        tag: str
            This is the XML element name (or tag name).

        Returns
        -------
        node: XMLTree
            An XML branch tree starting from `node`.
        '''

        assert isinstance(tag, str), 'tag invalid'

        node = self.__xml_tree_node.find(tag)

        assert node is not None

        node = XMLTree(node)  # wrap Element

        return node

    def get_all_sub_nodes(self, tag):
        '''
        Returns a list of all direct children nodes in the element tree with tag name.

        Parameters
        ----------
        tag: `str`
            XML element name or tag name, *e.g.*: `<task></task>`

        Returns
        -------
        sub_nodes: list(XMLTree)
            List of XMLTree items.
        '''

        assert isinstance(tag, str), 'tag invalid'

        sub_nodes = self.__xml_tree_node.findall(tag)

        assert isinstance(sub_nodes,list)
        assert len(sub_nodes) > 0

        xml_tree_list = [XMLTree(sub_node) for sub_node in sub_nodes]

        return xml_tree_list

    def get_node_children(self):
        '''
        Returns a list of the direct sub-elements in the given element (node) containing:
        the subnode, the tag (element) name, the attributes a list of tuples,
        and the content (text) of the node. This is not recursive. Recursion
        can be done by calling this method on the children nodes (that is the first
        element of the tuple).

        Parameters
        ----------
        empty:

        Returns
        -------
        children: list(tuple)
                  Tuple: (node, tag, [(attribute name,attribute value),(.,.)...],
                  content). Attribute name and value are string type.
        '''

        children = list()

        for child in self.__xml_tree_node:

            children.append( (child, child.tag, child.items(), child.text) )

        return children

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

#======================= end class XMLTree =======================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Cortix driver for the PyPlot module.
'''
# *********************************************************************************
import os
import sys
import io
import time
import datetime
import logging

from cortix.modulib.pyplot.pyplot import PyPlot
# *********************************************************************************

class CortixDriver():

    def __init__( self,
                  slot_id,
                  input_full_path_file_name,
                  manifest_full_path_file_name,
                  work_dir,
                  ports=list(),
                  cortix_start_time=0.0,
                  cortix_final_time=0.0,  # start time + evolution time
                  cortix_time_step=0.0,
                  cortix_time_unit=None
                ):

        # Sanity test
        assert isinstance(
            slot_id, int), '-> slot_id type %r is invalid.' % type(slot_id)
        assert isinstance(
            ports, list), '-> ports type %r is invalid.' % type(ports)
        assert len(ports) > 0
        assert isinstance(cortix_start_time, float), '-> time type %r is invalid.' % \
            type(cortix_start_time)
        assert isinstance(cortix_final_time, float), '-> time type %r is invalid.' % \
            type(cortix_final_time)
        assert isinstance(cortix_time_step, float), '-> time step type %r is invalid.' % \
            type(cortix_time_step)
        assert isinstance(cortix_time_unit, str), '-> time unit type %r is invalid.' % \
            type(cortix_time_unit)

        # Logging
        self.__log = logging.getLogger('launcher-modulib.pyplot_' + str(slot_id) +\
                                      '.cortix_driver')
        self.__log.debug('initializing an object of CortixDriver()')

        self.pyplot = PyPlot( slot_id,
                              input_full_path_file_name,
                              manifest_full_path_file_name,
                              work_dir,
                              ports,
                              cortix_start_time,
                              cortix_final_time,
                              cortix_time_step,
                              cortix_time_unit )

        self.__time_stamp = None  # temporary

        return
# ---------------------- end def __init__():------------------------------

    def call_ports(self, cortix_time=0.0):
        '''
        Call all ports at cortix_time
        '''

        self.__log_debug(cortix_time, 'call_ports')

        self.pyplot.call_ports(cortix_time)

        self.__log_debug(cortix_time, 'call_ports')

        return
# ---------------------- end def call_ports():----------------------------

    def execute(self, cortix_time=0.0, time_step=0.0):
        '''
        Evolve system from cortix_time to cortix_time + time_step
        '''

        self.__log_debug(cortix_time, 'execute')

        self.pyplot.execute(cortix_time, time_step)

        self.__log_debug(cortix_time, 'execute')

        return
# ---------------------- end def execute():-------------------------------

# *********************************************************************************
# Private helper functions (internal use: __)

    def __log_debug(self, cortix_time=0.0, caller='null-function-name'):

        if self.__time_stamp is None:
            s = ''
            self.__log.debug(s)
            s = '=========================================================='
            self.__log.debug(s)
            s = 'CORTIX::DRIVER->***->DRIVER->***->DRIVER->***->DRIVER->***'
            self.__log.debug(s)
            s = '=========================================================='
            self.__log.debug(s)

            s = caller + '(' + str(round(cortix_time, 2)) + '[min]):'
            self.__log.debug(s)

            self.__time_stamp = time.time()

        else:

            end_time = time.time()

            s = caller + '(' + str(round(cortix_time, 2)) + '[min]): '
            m = 'CPU elapsed time (s): ' + \
                str(round(end_time - self.__time_stamp, 2))
            self.__log.debug(s + m)

            self.__time_stamp = None
            if caller == 'execute':
                s = ''
                self.__log.debug(s)

        return
# ---------------------- end def __log_debug():---------------------------

# ====================== end class CortixDriver: =========================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Wind module example in Cortix.
'''
#*********************************************************************************
import os, sys, io, time
import logging
from collections             import namedtuple
import numpy as np

from cortix.src.utils.xmltree import XMLTree
from cortix.support.quantity import Quantity
from cortix.support.specie   import Specie
from cortix.support.phase    import Phase
#*********************************************************************************

class Wind():
    r'''
    Wind module used example in Cortix.
    '''

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__( self, slot_id,
                  input_full_path_file_name,
                  manifest_full_path_file_name,
                  work_dir, ports   = list(),
                  cortix_start_time = 0.0,
                  cortix_final_time = 0.0,
                  cortix_time_step  = 0.0,
                  cortix_time_unit  = None ):
        #.........................................................................
        # Sanity test 
        assert isinstance(slot_id, int),'-> slot_id type %r is invalid.'%type(slot_id)
        assert isinstance(ports, list),'-> ports type %r is invalid.'%type(ports)
        assert len(ports) > 0
        assert isinstance(cortix_start_time,float),'-> time type %r is invalid.'%\
               type(cortix_start_time)
        assert isinstance(cortix_final_time, float),'-> time type %r is invalid.'%\
               type(cortix_final_time)
        assert isinstance(cortix_time_step, float),'-> time step type %r is invalid.'%\
               type(cortix_time_step)
        assert isinstance(cortix_time_unit, str),'-> time unit type %r is invalid.'%\
               type(cortix_time_unit)

        # Logging: access Cortix Launcher logging facility
        self.__log = logging.getLogger('launcher-wind_'+str(slot_id)+'.cortix_driver.wind')
        self.__log.info('initializing an object of Wind()')

       # Read the manisfest
        self.__read_manifest( manifest_full_path_file_name )
        self.__log.info(self.__port_diagram)

        #.........................................................................
        # Member data 

        self.__slot_id = slot_id
        self.__ports   = ports

        # Convert Cortix's time unit to Wind's internal time unit
        if cortix_time_unit == 'minute':
           self.__time_unit_scale = 60.0
        elif cortix_time_unit == 'second':
           self.__time_unit_scale = 1.0
        elif cortix_time_unit == 'hour':
           self.__time_unit_scale = 60.0*60.0
        else:
           assert False, 'Cortix time_unit: %r not acceptable.' % time_unit

        self.__cortix_time_unit = cortix_time_unit

        self.__start_time = cortix_start_time * self.__time_unit_scale # Wind time unit
        self.__final_time = cortix_final_time * self.__time_unit_scale # Wind time unit

        if work_dir[-1] != '/': work_dir = work_dir + '/'
        self.__wrkDir = work_dir

        # Signal to start operation
        self.__goSignal = True     # start operation immediately
        for port in self.__ports:  # if there is a signal port, start operation accordingly
            (port_name, port_type, this_port_file) = port
            if port_name == 'go-signal' and port_type == 'use': self.__go_signal = False

        self.__setup_time = 60.0  # time unit; a delay time before starting to run

        #.........................................................................
        # Read input file information if any

        #fin = open(input_full_path_file_name,'r')

        # Configuration member data   

        self.__pyplot_scale = 'linear'
        #self.__ode_integrator = 'scikits.odes' # or 'scipy.integrate' 
        self.__ode_integrator = 'scipy.integrate'

        # Domain specific member data 

        shear_coeff = 0.4  # dimensionless
        Params = namedtuple('Params',['shear_coeff'])
        self.__params = Params( shear_coeff )

        # Setup the material phase as a gas
        species = list()

        air = Specie( name='air', formulaName='Air', phase='gas' )
        air.massCCUnit = 'g/cc'
        air.molarCCUnit = 'mole/cc'
        air.molarMass = 0.3*16*2 + 0.7*14*2

        species.append(air)

        quantities = list()

        # altitude
        altitude = Quantity( name='altitude', formalName='Altitude', unit='m' )
        quantities.append( altitude )
        # velocity
        velocity = Quantity( name='velocity', formalName='Velocity', unit='m/s' )
        quantities.append( velocity )

        self.__gas_phase = Phase( self.__start_time, species=species, quantities=quantities)

        # Initialize phase
        air_mass_cc = 0.1 # [g/cc]
        self.__gas_phase.SetValue( 'air', air_mass_cc, self.__start_time )

        x_0 = 1000.0  # initial altitude [m] above ground at 0
        self.__gas_phase.SetValue( 'altitude', x_0, self.__start_time )

        v_0 = np.array([0.0,0.0,0.0])   # initial wind velocity [m/s]
        self.__gas_phase.SetValue( 'velocity', v_0, self.__start_time )

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def call_ports( self, cortix_time=0.0 ):
        '''
        Transfer data at cortix_time
        '''

        cortix_time *= self.__time_unit_scale  # convert to Wind time unit

        # provide data to all provide ports 
        self.__provide_data( provide_port_name='state',  at_time=cortix_time )
        self.__provide_data( provide_port_name='output', at_time=cortix_time )

        # use data using the 'use-port-name' of the module
        self.__use_data( use_port_name='spatial-position', at_time=cortix_time )

        return

    def execute( self, cortix_time=0.0, cortix_time_step=0.0 ):
        '''
        Evolve system from cortix_time to cortix_time + cortix_time_step
        '''

        cortix_time      *= self.__time_unit_scale  # convert to Wind time unit
        cortix_time_step *= self.__time_unit_scale  # convert to Wind time unit

        s = 'execute('+str(round(cortix_time,2))+'[min]): '
        self.__log.debug(s)

        self.__evolve( cortix_time, cortix_time_step )

        return

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __provide_data( self, provide_port_name=None, at_time=0.0 ):

        # Access the port file
        port_file = self.__get_port_file( provide_port_name = provide_port_name )

        # Provide data to port files
        if provide_port_name == 'output' and port_file is not None:
            self.__provide_output( port_file, at_time )

        if provide_port_name == 'state' and port_file is not None:
            self.__provide_state( port_file, at_time )

        if provide_port_name == 'wind_velocity' and port_file is not None:
            self.__provide_wind_velocity( port_file, at_time )

        return

    def __use_data( self, use_port_name=None, at_time=0.0 ):

        # Access the port file
        port_file = self.__get_port_file( use_port_name = use_port_name )

        # Use data from port file
        if use_port_name == 'altitude' and port_file is not None:
            self.__use_altitude( port_file, at_time )

        return

    def __get_port_file( self, use_port_name=None, provide_port_name=None ):
        '''
        This may return a None port_file
        '''

        port_file = None

        #..........
        # Use ports
        #..........
        if use_port_name is not None:

            assert provide_port_name is None

            for port in self.__ports:

                (port_name,port_type,this_port_file) = port

                if port_name == use_port_name and port_type == 'use':
                    port_file = this_port_file

            if port_file is None:
                return None

            max_n_trials = 50
            n_trials     = 0
            while os.path.isfile(port_file) is False and n_trials <= max_n_trials:
                n_trials += 1
                time.sleep(0.1)

            if n_trials > max_n_trials:
                s = '__get_port_file(): waited ' + str(n_trials) + ' trials for port: ' +\
                    port_file
                self.__log.warn(s)

            assert os.path.isfile(port_file) is True, \
                   'port_file %r not available; stop.' % port_file

        #..............
        # Provide ports
        #..............
        if provide_port_name is not None:

            assert use_port_name is None

            for port in self.__ports:
                (port_name,port_type,this_port_file) = port
                if port_name == provide_port_name and port_type == 'provide':
                    port_file = this_port_file

        return port_file

    def __provide_output( self, port_file, at_time ):
        '''
        Provide data that will remain in disk after the simulation ends.
        '''
        import datetime

        # if the first time step, write the header of a table data file
        if at_time == self.__start_time:

            assert os.path.isfile(port_file) is False, \
                'port_file %r exists; stop.' % port_file
            fout = open(port_file,'w')

            # write file header
            fout.write('# name:   '+'wind_'+str(self.__slot_id)); fout.write('\n')
            fout.write('# author: '+'cortix.examples.modulib.wind'); fout.write('\n')
            fout.write('# version:'+'0.1'); fout.write('\n')
            today = datetime.datetime.today()
            fout.write('# today:  '+str(today)); fout.write('\n')
            fout.write('#')
            fout.write('%17s'%('Time[sec]'))

            # mass concentration
            for specie in self.__gas_phase.GetSpecies():
                fout.write('%18s'%(specie.formulaName+'['+specie.massCCUnit+']'))
            # quantities     
            for quant in self.__gas_phase.GetQuantities():
                fout.write('%18s'%(quant.formalName+'['+quant.unit+']'))

            fout.write('\n')
            fout.close()

        # write history data
        fout = open(port_file,'a')

        # Wind time 
        fout.write('%18.6e' % (at_time))

        # mass density   
        for specie in self.__gas_phase.GetSpecies():
            rho = self.__gas_phase.GetValue(specie.name, at_time)
            fout.write('%18.6e'%(rho))
        # quantities     
        for quant in self.__gas_phase.GetQuantities():
            val = self.__gas_phase.GetValue(quant.name, at_time)
            if quant.name == 'velocity':
                for v in val:
                    fout.write('%18.6e'%(v))
            else:
                fout.write('%18.6e'%(val))

        fout.write('\n')
        fout.close()

        return

    def __provide_state( self, port_file, at_time ):
        '''
        Provide the internal state of the module. This is typically used by another
        module such as '.cortix.modulib.pyplot'; this XML file type is now deprecated.
        However to have a dynamic plotting option, create this history file in the
        time unit of Cortix; that is, undo the time scaling.
        '''

        import datetime
        import xml.etree.ElementTree as ElementTree
        from threading import Lock

        n_digits_precision = 8

        # write header
        if at_time == self.__start_time:

            assert os.path.isfile(port_file) is False, 'port_file %r exists; stop.'%port_file

            tree = ElementTree.ElementTree()
            root_node = tree.getroot()

            a = ElementTree.Element('time-sequence')
            a.set('name','wind_'+str(self.__slot_id)+'-state')

            b = ElementTree.SubElement(a,'comment')
            today = datetime.datetime.today()
            b.set('author','cortix.examples.modulib.wind')
            b.set('version','0.1')

            b = ElementTree.SubElement(a,'comment')
            today = datetime.datetime.today()
            b.set('today',str(today))

            b = ElementTree.SubElement(a,'time')
            b.set('unit',self.__cortix_time_unit)

            # setup the headers
            for specie in self.__gas_phase.species:
                b = ElementTree.SubElement(a,'var')
                formula_name = specie.formulaName
                b.set('name',formula_name)
                unit = specie.massCCUnit
                b.set('unit',unit)
                b.set('legend','Wind_'+str(self.__slot_id)+'-state')
                b.set('scale',self.__pyplot_scale)

            for quant in self.__gas_phase.quantities:
                b = ElementTree.SubElement(a,'var')
                formal_name = quant.formalName
                b.set('name',formal_name)
                unit = quant.unit
                b.set('unit',unit)
                b.set('legend','Wind_'+str(self.__slot_id)+'-state')
                b.set('scale',self.__pyplot_scale)

            # write values for all variables
            b = ElementTree.SubElement(a,'timeStamp')
            b.set('value',str(round(at_time/self.__time_unit_scale,n_digits_precision)))

            values = list()

            for specie in self.__gas_phase.species:
                val = self.__gas_phase.GetValue( specie.name, at_time )
                values.append( val )

            for quant in self.__gas_phase.quantities:
                val = self.__gas_phase.GetValue( quant.name, at_time )
                values.append( val )

            # flush out data
            text = str()
            for value in values:
                text += str(round(value,n_digits_precision)) + ','

            text = text[:-1]

            b.text = text

            tree = ElementTree.ElementTree(a)

            tree.write( port_file, xml_declaration=True, encoding="unicode", method="xml" )

        #-------------------------------------------------------------------------
        # if not the first time step then parse the existing history file and append
        else:

            mutex = Lock()
            mutex.acquire()

            tree = ElementTree.parse( port_file )
            root_node = tree.getroot()

            a = ElementTree.Element('timeStamp')
            a.set('value',str(round(at_time/self.__time_unit_scale,n_digits_precision)))

            # all variables values
            values = list()

            for specie in self.__gas_phase.species:
                val = self.__gas_phase.GetValue( specie.name, at_time )
                values.append( val )
            for quant in self.__gas_phase.quantities:
                val = self.__gas_phase.GetValue( quant.name, at_time )
                values.append( val )

            # flush out data
            text = str()
            for value in values:
                text += str(round(value,n_digits_precision)) + ','

            text = text[:-1]

            a.text = text

            root_node.append(a)

            tree.write( port_file, xml_declaration=True, encoding="unicode", method="xml" )

            mutex.release()

        return

    def __evolve( self, cortix_time=0.0, cortix_time_step=0.0 ):
        r'''
         '''

        altitude = self.__gas_phase.GetValue('altitude',cortix_time)
        velocity = self.__gas_phase.GetValue('velocity',cortix_time)

        new_altitude = altitude
        new_velocity = velocity

        values = self.__gas_phase.GetRow( cortix_time ) # values at previous time
        at_time = cortix_time + cortix_time_step
        self.__gas_phase.AddRow( at_time, values ) # repeat values for current time

        self.__gas_phase.SetValue( 'altitude', new_altitude, at_time )  # update current values
        self.__gas_phase.SetValue( 'velocity', new_velocity, at_time ) # update current values

        return

    def __read_manifest( self, xml_tree_file ):
        '''
        Parse the manifest
        '''

        assert isinstance(xml_tree_file, str)

        # Read the manifest 
        xml_tree = XMLTree( xml_tree_file=xml_tree_file )

        assert xml_tree.get_node_tag() == 'module_manifest'

        assert xml_tree.get_node_attribute('name') == 'wind'

        # List of (port_name, port_type, port_mode, port_multiplicity)
        __ports = list()

        self.__port_diagram = 'null-module-port-diagram'

        # Get manifest data  
        for child in xml_tree.get_node_children():
            (elem, tag, attributes, text) = child

            if tag == 'port':

                text = text.strip()

                assert len(attributes) == 3, "only <= 3 attributes allowed."

                tmp = dict()  # store port name and three attributes

                for attribute in attributes:
                    key = attribute[0].strip()
                    val = attribute[1].strip()

                    if key == 'type':
                        assert val == 'use' or val == 'provide' or val == 'input' or \
                            val == 'output', 'port attribute value invalid.'
                        tmp['port_name'] = text  # port_name
                        tmp['port_type'] = val   # port_type
                    elif key == 'mode':
                        file_value = val.split('.')[0]
                        assert file_value == 'file' or file_value == 'directory',\
                            'port attribute value invalid.'
                        tmp['port_mode'] = val
                    elif key == 'multiplicity':
                        tmp['port_multiplicity'] = int(val)  # port_multiplicity
                    else:
                        assert False, 'invalid port attribute: %r=%r. fatal.'%\
                                (key,val)

                assert len(tmp) == 4
                store = (tmp['port_name'], tmp['port_type'], tmp['port_mode'],
                         tmp['port_multiplicity'])

                # (port_name, port_type, port_mode, port_multiplicity)
                __ports.append(store)

                # clear
                tmp   = None
                store = None

            if tag == 'diagram':

                self.__port_diagram = text

            if tag == 'ascii_art':

                self.__ascii_art = text

        return

#======================= end class Wind: =========================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import os
from cortix import Cortix
#*********************************************************************************

def run():
    '''
    Run the Cortix Droplet example. If Cortix and its dependencies are installed,
    this program should be executed at the command prompt inside the directory
    this program resides, namely, cortix/cortix/example/console_run/ directory.
    At the end of execution a directory with all logging and outputs is left in the
    work directory, as specified in the cortix-config-droplet.xml file. In this case,
    /tmp/cortix-droplet-wrk/.
    '''

    pwd = os.path.dirname(__file__)
    full_path_config_file = os.path.join(pwd, '../input/cortix-config-droplet.xml')

    # Create the cortix object and name it so it matches the <name></name> element
    # in the XML config file.
    cortix = Cortix('cortix-droplet', full_path_config_file)

    # This is a test of the droplet module alone and it needs no input.
    cortix.run_simulations(task_name="solo-droplet")

    # This is a test of the pyplot module alone: it needs an input file: state.xml which
    # is provided in ../input/. The input file ../input/pyplot.input points to this
    # file at runtime. 
    cortix.run_simulations(task_name="solo-pyplot")

    # This is a test of the droplet and pyplot connected.
    cortix.run_simulations(task_name="droplet-fall")

    # WARNING: the tasks are interferring with each other and threads are left running.
    # note how the droplet-fall runs two droplet instances...FIX ME.

#*********************************************************************************
if __name__ == "__main__":
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
This is a simple way to hide the name of species of interest in a simulation.
The user would modify and copy this class into the Cortix module of interest
and keep it private.

Author: Valmor de Almeida dealmeidav@ornl.gov; vfda
Sat Aug 15 13:41:12 EDT 2015
"""
#*******************************************************************************
import os
import sys
#*******************************************************************************

class Actor():
 """
 See atoms list in Specie.
 """

 def __init__( self, name ):

  assert isinstance(name, str)

 # ** this is the section to be modified by the user of this class **
  self.__name_atoms_formula = {
        'water16'   : [['2*H-1','O-16'],'H2O'],
        'spc(v)'    : [['2*O-16'],'O2'],
        'spc1(v)'   : [['O-16'],'O'],
        'spc2(v)'   : [['Xe-136'],'Xe'],
        'spc3(v)'   : [['I-127'],'I'],
            }
 # ** do not modify beyond this point **

  assert name in self.__name_atoms_formula.keys(), 'name %r not valid.' % name

  self.__atoms = self.__name_atoms_formula[name][0]

  formula = self.__name_atoms_formula[name][1]
  assert isinstance(formula, str), 'formula %r not valid.' % formula
  self.__formula = formula
#----------------------- end def __init__():--------------------------------------

 def __get_atoms(self):
  return self.__atoms
 atoms = property(__get_atoms, None, None, None)

 def __get_formula(self):
  return self.__formula
 formula = property(__get_formula, None, None, None)

#======================= end class Launcher: =====================================




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
The Cortix class definition.

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import os
import logging
from cortix.src.simulation import Simulation
from cortix.src.utils.xmltree import XMLTree
from cortix.src.utils.set_logger_level import set_logger_level
#*********************************************************************************

class Cortix():
    '''
    The main Cortix class definition. This class encapsulates the
    concepts of simulations, tasks, and modules, for a given application providing
    the user with an interface to the simulations.
    '''

#*********************************************************************************
# Construction 
#*********************************************************************************

    def __init__(self, name, config_xml_file='cortix-config.xml'):

        assert isinstance(name,str), 'must give Cortix object a name'

        assert isinstance(config_xml_file, str), '-> config_xml_file not a str.'

        # Create the configuration XML tree
        config_xml_tree = XMLTree( xml_tree_file=config_xml_file )

        assert config_xml_tree.get_node_tag() == 'cortix_config'

        # Read the cortix config element (tag) name <name></name>
        node = config_xml_tree.get_sub_node('name') # get sub_node w/ tag: name (XML element)

        # Set the Cortix configuration name
        self.__name = node.get_node_content()  # name is now, say, 'droplet-fall'

        # Consistency check
        assert self.__name == name,\
            'Runtime Cortix object name %r conflicts with cortix-config.xml %r' \
            % (self.__name, name)

        # Read the work directory name
        node = config_xml_tree.get_sub_node('work_dir')
        work_dir = node.get_node_content()
        if work_dir[-1] != '/':
            work_dir += '/'

        self.__work_dir = work_dir + self.__name + '-wrk/'

        # Create the work directory
        if os.path.isdir(self.__work_dir):
            os.system('rm -rf ' + self.__work_dir)

        os.system('mkdir -p ' + self.__work_dir)

        # Create the logging facility for each object
        self.__create_logging( config_xml_tree )

        self.__log.info('Created Cortix work directory: %s', self.__work_dir)

        #==================
        # Setup simulations (one or more as specified in the config file)
        #==================
        self.__setup_simulations( config_xml_tree )

        self.__log.info('Created Cortix object %s', self.__name)

        return

    def __del__(self):

        self.__log.info("Destroyed Cortix object: %s", self.__name)

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def run_simulations(self, task_name=None):
        '''
        This method runs every simulation defined by the Cortix object. At the 
        moment this is done one simulation at a time.
        '''

        for sim in self.__simulations:

            sim.execute( task_name )

        return

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __create_logging(self, config_xml_tree):
        '''
        A helper function to setup the logging facility used in self.__init__()
        '''

        logger_name = self.__name

        self.__log = logging.getLogger(logger_name)
        self.__log.setLevel(logging.NOTSET)

        node = config_xml_tree.get_sub_node('logger') # tag name is logger

        logger_level = node.get_node_attribute('level')
        self.__log = set_logger_level(self.__log, logger_name, logger_level)

        file_handler = logging.FileHandler(self.__work_dir + 'cortix.log')
        file_handler.setLevel(logging.NOTSET)
        file_handler_level = None

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.NOTSET)
        console_handler_level = None

        for child in node.get_node_children():
            (elem,tag,attributes,text) = child
            elem = XMLTree( elem ) # fixme: remove wrapping
            if tag == 'file_handler':
                file_handler_level = elem.get_node_attribute('level')
                file_handler = set_logger_level(file_handler, logger_name,
                                                file_handler_level)
            if tag == 'console_handler':
                console_handler_level = elem.get_node_attribute('level')
                console_handler = set_logger_level(console_handler, logger_name,
                                                   console_handler_level)

        # Formatter added to handlers
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        console_handler.setFormatter(formatter)

        # add handlers to logger
        self.__log.addHandler(file_handler)
        self.__log.addHandler(console_handler)
        self.__log.info('Created Cortix logger: %s', self.__name)
        self.__log.debug('Logger level: %s', logger_level)
        self.__log.debug('Logger file handler level: %s', file_handler_level)
        self.__log.debug('Logger console handler level: %s', console_handler_level)

        return

    def __setup_simulations(self, config_xml_tree):
        '''
        This method is a helper function for the Cortix constructor whose purpose is 
        to set up the simulations defined by the Cortix configuration.
        '''

        self.__simulations = list()

        for sim_config_xml_node in config_xml_tree.get_all_sub_nodes('simulation'):

            assert sim_config_xml_node.get_node_tag() == 'simulation'

            self.__log.debug('__setup_simulations(): simulation name: %s',
                    sim_config_xml_node.get_node_attribute('name'))

            simulation = Simulation( self.__work_dir, sim_config_xml_node )

            self.__simulations.append(simulation)

        return

#======================= end class Cortix: =======================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
The Cortix Task class definition.

Cortix: a program for system-level modules coupling, execution, and analysis.
'''
#*********************************************************************************
import os
import time
import logging
from xml.etree.cElementTree import ElementTree
from cortix.src.utils.xmltree import XMLTree
from cortix.src.utils.set_logger_level import set_logger_level
#*********************************************************************************

class Task:
    '''
    A Task is work done by a Simulation handled by Cortix.
    A Task will use a given Application.
    '''

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self, parent_work_dir=None, task_config_node=XMLTree()):

        assert isinstance(parent_work_dir, str), '-> parent_work_dir invalid.'

        # Inherit a configuration tree
        assert isinstance(task_config_node, XMLTree), \
            '-> task_config_node not a XMLTree: %r.' % type(task_config_node)
        self.__config_node = task_config_node

        # Read the task name
        self.__name = self.__config_node.get_node_attribute('name')

        # Set the work directory (previously created)
        assert os.path.isdir(parent_work_dir), 'work directory not available.'
        self.__work_dir = parent_work_dir + 'task_' + self.__name + '/'
        os.system('mkdir -p ' + self.__work_dir)

        # Create the logging facility for the object
        node = task_config_node.get_sub_node('logger')
        logger_name = 'task:'+self.__name
        self.__log = logging.getLogger(logger_name)
        self.__log.setLevel(logging.NOTSET)

        logger_level = node.get_node_attribute('level')
        self.__log = set_logger_level(self.__log, logger_name, logger_level)

        file_handle = logging.FileHandler(self.__work_dir + 'task.log')
        file_handle.setLevel(logging.NOTSET)

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.NOTSET)

        for child in node.get_node_children():
            (elem,tag,attributes,text) = child
            elem = XMLTree( elem )
            if tag == 'file_handler':
                file_handle_level = elem.get_node_attribute('level')
                file_handle = set_logger_level(file_handle, logger_name,
                                               file_handle_level)
            if tag == 'console_handler':
                console_handle_level = elem.get_node_attribute('level')
                console_handler = set_logger_level(console_handler,
                                                   logger_name, console_handle_level)

        # formatter added to handlers
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        file_handle.setFormatter(formatter)
        console_handler.setFormatter(formatter)

        # add handlers to logger
        self.__log.addHandler(file_handle)
        self.__log.addHandler(console_handler)

        self.__log.info('created logger: %s', self.__name)
        self.__log.debug('logger level: %s', logger_level)
        self.__log.debug('logger file handler level: %s', file_handle_level)
        self.__log.debug('logger console handler level: %s', console_handle_level)

        self.__start_time = self.__evolve_time = self.__time_step = 0.0
        self.__start_time_unit = 'null-start_time_unit'
        self.__evolve_time_unit = 'null-evolve_time_unit'
        self.__time_step_unit = 'null-time_step_unit'
        self.__runtime_cortix_param_file = 'null-runtime_cortix_param_file'

        self.__log.debug('start __init__()')
        for child in self.__config_node.get_node_children():
            (elem, tag, items, text) = child
            if tag == 'start_time':
                for (key, value) in items:
                    if key == 'unit':
                        self.__start_time_unit = value
                self.__start_time = float(text.strip())

            if tag == 'evolve_time':
                for (key, value) in items:
                    if key == 'unit':
                        self.__evolve_time_unit = value
                self.__evolve_time = float(text.strip())

            if tag == 'time_step':
                for (key, value) in items:
                    if key == 'unit':
                        self.__time_step_unit = value
                self.__time_step = float(text.strip())

        if self.__start_time_unit == 'null-start_time_unit':
            self.__start_time_unit = self.__evolve_time_unit
        assert self.__evolve_time_unit != 'null-evolve_time_unit', \
            'invalid time unit = %r' % (self.__evolve_time_unit)

        self.__log.debug('start_time value = %s', str(self.__start_time))
        self.__log.debug('start_time unit  = %s', str(self.__start_time_unit))
        self.__log.debug('evolve_time value = %s', str(self.__evolve_time))
        self.__log.debug('evolve_time unit  = %s', str(self.__evolve_time_unit))
        self.__log.debug('end __init__()')
        self.__log.info('created task: %s', self.__name)

        return

    def __del__(self):

        self.__log.info('destroyed task: %s', self.__name)

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def execute(self, application):
        '''
        This method is used to execute (accomplish) the given task.
        '''

        network = application.get_network(self.__name)

        runtime_status_files = dict()

        for slot_name in network.module_slot_names:

            module_name = '_'.join( slot_name.split('_')[:-1] )
            slot_id = int(slot_name.split('_')[-1])

            module = application.get_module(module_name)

            param_file = self.__runtime_cortix_param_file
            comm_file  = network.get_runtime_cortix_comm_file_name(slot_name)

            # Run module in the slot
            self.__log.info('call execute on module: %s:%s',module_name,slot_id)

            status_file = module.execute(slot_id, param_file, comm_file)

            assert status_file is not None, \
                   'module launching failed; module name: %r' % module_name

            runtime_status_files[slot_name] = status_file

        # monitor runtime status
        status = 'running'
        while status == 'running':
            time.sleep(5)  # hard coded; fix me.
            (status, slot_names) = self.__get_runtime_status(runtime_status_files)
            self.__log.info('runtime status: %s; module slots running: %s', status,
                          str(slot_names))

        return

    def __get_name(self):
        '''
        `str`:Task name
        '''

        return self.__name

    name = property(__get_name, None, None, None)

    def __get_work_dir(self):
        '''
        `str`:Working directory of task
        '''

        return self.__work_dir

    work_dir = property(__get_work_dir, None, None, None)

    def __get_start_time(self):
        '''
        `float`:Task initial time

        Parameters
        ---------
        empty

        Returns
        -------
        __start_time

        '''

        return self.__start_time

    start_time = property(__get_start_time, None, None, None)

    def __get_start_time_unit(self):
        '''
        `str`:Task initial time unit
        '''

        return self.__start_time_unit

    start_time_unit = property(__get_start_time_unit, None, None, None)

    def __get_evolve_time(self):
        '''
        `float`:Task final time
        '''

        return self.__evolve_time

    evolve_time = property(__get_evolve_time, None, None, None)

    def __get_evolve_time_unit(self):
        '''
        `str`:Task final time unit
        '''

        return self.__evolve_time_unit

    evolve_time_unit = property(__get_evolve_time_unit, None, None, None)

    def __get_time_step(self):
        '''
        `float`:Magnitude of incremental step in the task's time
        '''

        return self.__time_step

    time_step = property(__get_time_step, None, None, None)

    def __get_time_step_unit(self):
        '''
        `str`:Time step unit
        '''

        return self.__time_step_unit

    time_step_unit = property(__get_time_step_unit, None, None, None)

    def set_runtime_cortix_param_file(self, full_path):
        '''
        Sets the task config file to the specified file.
        '''

        self.__runtime_cortix_param_file = full_path

        return

    def __get_runtime_cortix_param_file(self):
        '''
        `str`:Task's config file
        '''

        return self.__runtime_cortix_param_file

    runtime_cortix_param_file = property(__get_runtime_cortix_param_file,
            set_runtime_cortix_param_file, None, None)

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __get_runtime_status(self, runtime_status_files):
        '''
        Helper funcion for montioring the status of the task. It reads the status
        files of all modules and reports which ones are still running.
        '''

        task_status = 'finished'
        running_module_slots = list()

        for (slot_name, status_file) in runtime_status_files.items():
            if not os.path.isfile(status_file):
                time.sleep(0.1)

            assert os.path.isfile(status_file), 'runtime status file %r not found.' \
                % status_file

            tree = ElementTree()
            tree.parse(status_file)
            status_file_xml_root_node = tree.getroot()
            node = status_file_xml_root_node.find('status')

            status = node.text.strip()

            if status == 'running':
                task_status = 'running'
                running_module_slots.append(slot_name)

        return (task_status, running_module_slots)

#======================= end class Task: =========================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda

This Specie class is to be used with other classes in plant-level process modules.

NB: Species is always used either in singular or plural cases, the class
    named here reflects one species. If many species are used in an external
    context, the species object name can be used without conflict.

For unit testing do at the linux command prompt:
    python specie.py

NB: The Specie() class encapsulates either the molecular or empirical chemical
    formula of a compound.
    The definition of a chemical species here is extended to ficticious compounds.
    This is done as follows. Say MAO2 is either a molecular or empirical chemical
    formula of a ficticious compound denoting minor actinides dioxide. The list
    of atoms is given as follows:

    ['0.49*Np-237', '0.42*Am-241', '0.08*Am-243', '0.01*Cm-244', '2.0*O-16']

    note the MA forming nuclides add to 1 = 0.49 + 0.42 + 0.08 + 0.01. Therefore
    the number of atoms in this compound is 3. 1 MA "atom" and 2 O.
    Note that the total number of "atoms" is obtained by summing all multipliers:
    0.49 + 0.42 + 0.08 + 0.01 + 2.0.
    The nuclide is indicated by the element symbol followed by a dash and the
    atomic mass number. Here the number of nuclide types is 5 (self._nNuclideTypes).

    The numbers preceeding the nuclide symbol before the * will be referred to as
    multipliers. The sum of the multipliers will add to the number of "atoms" in
    the formula. WARNING: a multiplier could be in the format 0.00e-00. In this
    case a hiphen may appear twice, e.g.: 1.549e-09*U-233

    Other forms can be used for common true species

    ['Np-237', '2.0*O-16'] or ['Np-237', 'O-16', 'O-16'] or [ '2*H', 'O' ] or
    [ 'H', 'O', 'H' ]  etc...

    This code will calculate the molar mass of any species with a given valid
    atom list using a provided periodic table of chemical elements. The user
    can also reset the value of the molar mass with a setter method.

Sat May  9 21:40:48 EDT 2015 created; vfda
"""

# *******************************************************************************
import os
import sys

from cortix.support.periodictable import ELEMENTS
# *******************************************************************************

# *******************************************************************************


class Specie():

    # todo: phase should not be here; concentrations should not be here
    #      only molar quantities should be here
    #      see the Phase container

    def __init__(self,
                 name='null',
                 formulaName='null',
                 phase='null',
                 atoms=list(),
                 molarCC=0.0,      # default unit: M (mole/L)
                 massCC=0.0,      # default unit: g/L
                 flag=None):

        assert isinstance(name, str), 'oops not string.'
        self._name = name

        assert isinstance(formulaName, str), 'oops not string.'
        self._formulaName = formulaName

        assert isinstance(phase, str), 'oops not string.'
        self._phase = phase

        assert isinstance(atoms, list), 'oops not list.'
        self._atoms = atoms

        self._flag = flag  # flag can be any type

        self._molarMass = 0.0
        self._molarHeatPwr = 0.0
        self._molarGammaPwr = 0.0
        self._molarRadioactivity = 0.0

        self._molarMassUnit = 'g/mole'

        self._molarHeatPwrUnit = 'W/mole'
        self._molarGammaPwrUnit = 'W/mole'
        self._molarRadioactivityUnit = 'Ci/mole'

        self._molarRadioactivityFractions = list()

        self._molarCCUnit = 'mole/L'
        self._massCCUnit = 'g/L'

        self.__UpdateMolarMass()

        if self._molarMass == 0.0:
            self._molarCC = 0.0
            self._massCC = 0.0
            return

        assert isinstance(molarCC, float), 'oops not a float.'
        assert molarCC >= 0.0, 'oops negative value.'
        self._molarCC = molarCC
        self._massCC = molarCC * self._molarMass

        assert isinstance(massCC, float), 'oops not a float.'
        assert massCC >= 0.0, 'oops negative value.'
        if self._massCC == 0.0:
            self._massCC = massCC
            self._molarCC = massCC / self._molarMass

        return


# *******************************************************************************

# *******************************************************************************
# Setters and Getters methods
# -------------------------------------------------------------------------------
# These are passing arguments by value effectively. Because the python objects
# passed into/out of the function are immutable.

    def GetName(self):
        return self._name

    def SetName(self, n):
        self._name = n
    name = property(GetName, SetName, None, None)

    def GetFormulaName(self):
        return self._formulaName

    def SetFormulaName(self, f):
        self._formulaName = f
    formulaName = property(GetFormulaName, SetFormulaName, None, None)

    def GetPhase(self):
        return self._phase

    def SetPhase(self, p):
        self._mass = p
    phase = property(GetPhase, SetPhase, None, None)

    def GetMolarMass(self):
        return self._molarMass

    def SetMolarMass(self, v):
        self._molarMass = v
    molarMass = property(GetMolarMass, SetMolarMass, None, None)

    def GetMolarMassUnit(self):
        return self._molarMassUnit

    def SetMolarMassUnit(self, v):
        self._molarMassUnit = v
    molarMassUnit = property(GetMolarMassUnit, SetMolarMassUnit, None, None)

    def GetMolarRadioactivity(self):
        return self._molarRadioactivity

    def SetMolarRadioactivity(self, v):
        self._molarRadioactivity = v
    molarRadioactivity = property(
        GetMolarRadioactivity,
        SetMolarRadioactivity,
        None,
        None)

    def GetMolarRadioactivityFractions(self):
        return self._molarRadioactivityFractions

    def SetMolarRadioactivityFractions(self, fracs):
        assert isinstance(fracs, list), 'oops not list.'
        if len(fracs) > 0:
            assert len(fracs) == len(self._atoms), 'oops not right length,'
        if len(fracs) != 0:
            assert isinstance(fracs[-1], float), 'oops not float.'
        self._molarRadioactivityFractions = fracs
    molarRadioactivityFractions = property(
        GetMolarRadioactivityFractions,
        SetMolarRadioactivityFractions,
        None,
        None)

    def GetMolarRadioactivityUnit(self):
        return self._molarRadioactivityUnit

    def SetMolarRadioactivityUnit(self, v):
        self._molarRadioactivityUnit = v
    molarRadioactivityUnit = property(
        GetMolarRadioactivityUnit,
        SetMolarRadioactivityUnit,
        None,
        None)

    def GetMolarHeatPwr(self):
        return self._molarHeatPwr

    def SetMolarHeatPwr(self, v):
        self._molarHeatPwr = v
    molarHeatPwr = property(GetMolarHeatPwr, SetMolarHeatPwr, None, None)

    def GetMolarHeatPwrUnit(self):
        return self._molarHeatPwrUnit

    def SetMolarHeatPwrUnit(self, v):
        self._molarHeatPwrUnit = v
    molarHeatPwrUnit = property(
        GetMolarHeatPwrUnit,
        SetMolarHeatPwrUnit,
        None,
        None)

    def GetMolarGammaPwr(self):
        return self._molarGammaPwr

    def SetMolarGammaPwr(self, v):
        self._molarGammaPwr = v
    molarGammaPwr = property(GetMolarGammaPwr, SetMolarGammaPwr, None, None)

    def GetMolarGammaPwrUnit(self):
        return self._molarGammaPwrUnit

    def SetMolarGammaPwrUnit(self, v):
        self._molarGammaPwrUnit = v
    molarGammaPwrUnit = property(
        GetMolarGammaPwrUnit,
        SetMolarGammaPwrUnit,
        None,
        None)

    # Deprecated; see new interface below as GetFormula
    def GetAtoms(self):
        return self._atoms

    def SetAtoms(self, atoms):
        assert isinstance(atoms, list), 'oops not list.'
        if len(atoms) != 0:
            assert isinstance(atoms[-1], str), 'oops not string.'
        self._atoms = atoms
        self.__UpdateMolarMass()
    atoms = property(GetAtoms, SetAtoms, None, None)

    # New interface
    def GetFormula(self):
        return self._atoms

    def SetFormula(self, atoms):
        assert isinstance(atoms, list), 'oops not list.'
        if len(atoms) != 0:
            assert isinstance(atoms[-1], str), 'oops not string.'
        self._atoms = atoms
        self.__UpdateMolarMass()
    formula = property(GetFormula, SetFormula, None, None)

    def GetNAtoms(self):  # number of ficticious atoms in the species (see NB above)
        return self._nAtoms
    nAtoms = property(GetNAtoms, None, None, None)

    # number of nuclide types involved in the species definition
    def GetNNuclideTypes(self):
        return self._nNuclideTypes
    nNuclideTypes = property(GetNNuclideTypes, None, None, None)

    def SetFlag(self, f):
        self._flag = f

    def GetFlag(self):
        return self._flag
    flag = property(GetFlag, SetFlag, None, None)

    def GetMolarCC(self):
        return self._molarCC

    def SetMolarCC(self, v):
        self._molarCC = v
        self._massCC = v * self._molarMass
    molarCC = property(GetMolarCC, SetMolarCC, None, None)

    def GetMolarCCUnit(self):
        return self._molarCCUnit

    def SetMolarCCUnit(self, v):
        self._molarCCUnit = v
    molarCCUnit = property(GetMolarCCUnit, SetMolarCCUnit, None, None)

    def GetMassCC(self):
        return self._massCC

    def SetMassCC(self, v):
        self._massCC = v
        if self._molarMass == 0.0 and v == 0.0:
            self._molarCC = 0.0
        else:
            self._molarCC = v / self._molarMass
    massCC = property(GetMassCC, SetMassCC, None, None)

    def GetMassCCUnit(self):
        return self._massCCUnit

    def SetMassCCUnit(self, v):
        self._massCCUnit = v
    massCCUnit = property(GetMassCCUnit, SetMassCCUnit, None, None)

# *******************************************************************************
# Internal helpers

    def __UpdateMolarMass(self):

        if len(self._atoms) == 0:
            return

        for entry in self._atoms:
            assert isinstance(entry, str), 'oops'
            tmp = entry.split('*')
            nuclide = tmp[-1]
            element = nuclide.split('-')[0]
            assert element in ELEMENTS, 'element = %r' % (element)

        self._nAtoms = 0
        self._nNuclideTypes = 0

        nuclides = dict()

        nAtoms = 0
        summ = 0.0
        for entry in self._atoms:
            assert isinstance(entry, str), 'oops'
            # format example:  3.2*O-18, or 3*O or O or O-16
            tmp = entry.split('*')
            multiple = 1.0
            # single nuclide
            if len(tmp) == 1:
                nuclide = tmp[0]
            # multiple nuclide
            elif len(tmp) == 2:
                multiple = float(tmp[0])
                nuclide = tmp[1]
            else:
                assert False

            nuclides[nuclide] = multiple
            nAtoms += multiple

            try:
                tmp = nuclide.split('-')
                if len(tmp) == 1:
                    element = ELEMENTS[tmp[0]]
                    molarMass = element.exactmass  # from isotopic composition
                    if molarMass == 0.0:
                        molarMass = element.mass
                elif len(tmp) == 2:
                    element = ELEMENTS[tmp[0]].isotopes[int(tmp[1].strip('m'))]
                    molarMass = element.mass
                else:
                    assert False
            except KeyError:
                summ += multiple * 0.0
            else:
                summ += multiple * molarMass

        self._molarMass = summ
#   print( summ )
        self._nAtoms = nAtoms
        self._nNuclideTypes = len(nuclides)

        return

    def __ReorderFormula(self):

        atoms1 = self._atoms[:]  # shallow copy
        atoms2 = list()

        if len(self._atoms) <= 1:
            return atoms1

        if len(self._atoms) > 1:

            # save the multiplier value as a string type of scientific notation
            for entry in self._atoms:

                assert isinstance(entry, str), 'oops'

                # format example:  3.2*O-18, or 3*O or O or O-16
                tmp = entry.split('*')

                multiplier = 0.0

                if len(tmp) == 1:
                    continue
                elif len(tmp) == 2:
                    multiplier = float(tmp[0])
                else:
                    assert False

                assert multiplier != 0.0, 'multiplier = %r' % (multiplier)

                multiplier = '{0:9.3e}'.format(multiplier)

                atoms1[self._atoms.index(entry)] = multiplier + '*' + tmp[1]

            # order in decreasing order of multiplier magnitude
            multipliers_lst = list()

            for entry in atoms1:

                tmp = entry.split('*')

                multiplier = 0.0

                if len(tmp) == 1:
                    continue
                elif len(tmp) == 2:
                    multiplier = float(tmp[0])
                else:
                    assert False

                multipliers_lst.append(float(multiplier))

            sortedAtoms_lst = [a for (i, a) in sorted(zip(multipliers_lst, atoms1),
                                                      key=lambda pair: pair[0],
                                                      reverse=True)]

            atoms2 = sortedAtoms_lst

        return atoms2


# *******************************************************************************
# Printing of data members

    def __str__(self):
        s = '\n\t Specie(): name=%s;' + ' formulaName=%s;' + ' phase=%s;' + '\n\t formula=%s;' + '\n\t # atoms=%s;' + ' # nuclide types=%s;' + ' molar mass=%9.3e[%s];' + ' molar cc=%9.3e[%s];' + ' mass cc=%9.3e[%s];' + '\n\t flag=%s;' + '\n\t molar radioactivity=%9.3e[%s];' + \
            '\n\t radioactivity  dens.=%9.3e[%s];' + '\n\t molar heat pwr=%9.3e[%s];' + '\n\t heat pwr dens.=%9.3e[%s];' + \
            '\n\t molar gamma pwr=%9.3e[%s];' + '\n\t gamma pwr dens.=%9.3e[%s];' + \
            '\n\t atoms=%s;' + '\n\t molar radioactivity fractions=%s'
        return s % (self.name, self.formulaName, self.phase, self.__ReorderFormula(), self.nAtoms, self.nNuclideTypes, self.molarMass, self.molarMassUnit, self.molarCC, self.molarCCUnit, self.massCC, self.massCCUnit, self.flag, self.molarRadioactivity, self.molarRadioactivityUnit, self.molarRadioactivity *
                    self.molarCC, '[Ci/cc]', self.molarHeatPwr, self.molarHeatPwrUnit, self.molarHeatPwr * self.molarCC, '[W/cc]', self.molarGammaPwr, self.molarGammaPwrUnit, self.molarGammaPwr * self.molarCC, '[W/cc]', [i.split('*')[-1] for i in self.formula], ['%9.3e' % i for i in self.molarRadioactivityFractions])

    def __repr__(self):
        s = '\n\t Specie(): name=%s;' + ' formulaName=%s;' + ' phase=%s;' + '\n\t formula=%s;' + '\n\t # atoms=%s;' + ' # nuclide types=%s;' + ' molar mass=%9.3e[%s];' + ' molar cc=%9.3e[%s];' + ' mass cc=%9.3e[%s];' + '\n\t flag=%s;' + '\n\t molar radioactivity=%9.3e[%s];' + \
            '\n\t radioactivity  dens.=%9.3e[%s];' + '\n\t molar heat pwr=%9.3e[%s];' + '\n\t heat pwr dens.=%9.3e[%s];' + \
            '\n\t molar gamma pwr=%9.3e[%s];' + '\n\t gamma pwr dens.=%9.3e[%s];' + \
            '\n\t atoms=%s;' + '\n\t molar radioactivity fractions=%s'
        return s % (self.name, self.formulaName, self.phase, self.__ReorderFormula(), self.nAtoms, self.nNuclideTypes, self.molarMass, self.molarMassUnit, self.molarCC, self.molarCCUnit, self.massCC, self.massCCUnit, self.flag, self.molarRadioactivity, self.molarRadioactivityUnit, self.molarRadioactivity *
                    self.molarCC, '[Ci/cc]', self.molarHeatPwr, self.molarHeatPwrUnit, self.molarHeatPwr * self.molarCC, '[W/cc]', self.molarGammaPwr, self.molarGammaPwrUnit, self.molarGammaPwr * self.molarCC, '[W/cc]', [i.split('*')[-1] for i in self.formula], ['%9.3e' % i for i in self.molarRadioactivityFractions])
# *******************************************************************************



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt

import os


def test_cortix_config_exists():
    '''
    Test to make sure a configuration file exists
    in the input directory.
    '''

    assert os.path.isfile("input/cortix-config-pyplot.xml")
    assert os.path.isfile("input/cortix-config-droplet.xml")

if __name__ == "__main__":
    test_cortix_config_exists()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
Phase *history* container. When you think of a phase value, think of that value at
a specific point in time.

ATTENTION:
The species (list of Specie) AND quantities (list of Quantity) data members
have ARBITRARY density values either at an arbitrary point in the history or at
no point in the history. This needs to be removed in the future to avoid confusion.

To obtain history values, associated to the phase, at a particular point in time,
use the GetValue() method to access the history data frame (pandas) via columns and
rows. The corresponding values in species and quantities are OVERRIDEN and NOT to
be used through the phase interface.

Author: Valmor F. de Almeida dealmeidav@ornl.gov; vfda
Sat Sep  5 01:26:53 EDT 2015
'''
#*********************************************************************************
import os
import sys
from copy import deepcopy
import numpy as npy
import pandas

from cortix.support.specie   import Specie
from cortix.support.quantity import Quantity
#*********************************************************************************

class Phase():

    def __init__(self,
                 time_stamp = None,
                 species    = None,
                 quantities = None
                 ):

        if time_stamp is None:
            time_stamp = 0.0
        assert isinstance(time_stamp, float)

        if species is not None:
            assert isinstance(species, list)
            for specie in species:
                assert isinstance(specie, Specie)

        if quantities is not None:
            assert isinstance(quantities, list)
            for quant in quantities: 
                assert isinstance(quant, Quantity)

# List of species and quantities objects; columns of data frame are named
# by objects
        # a new object held by a Phase() object
        self.__species = deepcopy(species)
        # a new object held by a Phase() object
        self.__quantities = deepcopy(quantities)

        names = list()

        if species is not None:
            for specie in self.__species:
                names.append(specie.name)
                specie.massCC = 0.0  # clear these values
                                     # todo: eliminate them from Specie in the future

        if quantities is not None:
            for quant in self.__quantities:
                names.append(quant.name)
                quant.value = 0.0    # clear these values 
                                     # todo: eliminate them from Quantity in the future

# Table data phase without data type assigned; this is left to the user
# Time stamps will always be float
        self.__phase = pandas.DataFrame( index=[float(time_stamp)], columns=names )
#        self.__phase.fillna( 0.0, inplace=True )  # dtype defaults to float

        return

# *******************************************************************************
# Setters and Getters methods
# -------------------------------------------------------------------------------
# These are passing arguments by value effectively. Because the python objects
# passed into/out of the function are immutable.

    def GetActors(self):
        return list(self.__phase.columns)  # return all names in order

    def GetTimeStamps(self):
        return list(self.__phase.index)  # return all time stamps
    timeStamps = property(GetTimeStamps, None, None, None)

    def GetSpecie(self, name):
        for specie in self.__species:
            if specie.name == name:
                time_stamp = self.__get_time_stamp( None ) # get latest time stamp
                assert name in self.__phase.columns, 'name %r not in %r'% \
                   (name,self.__phase.columns)
                specie.massCC = self.__phase.loc[time_stamp, name]
                return specie  # return specie syncronized with the phase
        return None

    def GetSpecies(self):
        for species in self.__species:
          tmp = self.GetSpecie(species.name) # handy way to synchronize the whole list
        return self.__species
    species = property(GetSpecies, None, None, None)

    def GetQuantities(self):
        for quant in self.__quantities:
          tmp = self.GetQuantity(quant.name) # handy way to synchronize the whole list
        return self.__quantities
    quantities = property(GetQuantities, None, None, None)

    def SetSpecieId(self, name, val):
        for specie in self.__species:
            if specie.name == name:
                specie.flag = val
                return

    def GetQuantity(self, name):
        for quant in self.__quantities:
            if quant.name == name:
                time_stamp = self.__get_time_stamp( None ) # get latest time stamp
                assert name in self.__phase.columns, 'name %r not in %r'% \
                   (name,self.__phase.columns)
                quant.value = self.__phase.loc[time_stamp, name]
                return quant  # return quantity syncronized with the phase
        return None

    def AddSpecie(self, new_specie):
        assert isinstance(new_specie, Specie)
        assert new_specie.name not in list(self.__phase.columns), \
               'new_specie: %r exists. Current names: %r' % \
               (new_specie, self.__phase.columns)
        speciesFormulae = [specie.formulaName for specie in self.__species]
        assert new_specie.formulaName not in speciesFormulae
        self.__species.append(new_specie)
        newName = new_specie.name
        col = pandas.DataFrame( index=list(self.__phase.index), columns=[newName] )
        tmp = self.__phase
        df = tmp.join(col, how='outer')
        self.__phase = df.fillna(0.0)   # for species have float as default

    def AddQuantity(self, newQuant):
        assert isinstance(newQuant, Quantity)
        assert newQuant.name not in list(self.__phase.columns), \
               'quantity: %r exists. Current names: %r' % \
               (newQuant, self.__phase.columns)
        quantFormalNames = [quant.formalName for quant in self.__quantities]
        assert newQuant.formalName not in quantFormalNames
        self.__quantities.append(newQuant)
        newName = newQuant.name

        # create a col with object data type; user must fill out column 
        col = pandas.DataFrame( index=list( self.__phase.index), columns=[newName],
                                dtype=object )
        tmp = self.__phase
        df  = tmp.join(col, how='outer')
#        self.__phase = df.fillna(newQuant.value)

    def AddRow(self, try_time_stamp, row_values):
        assert isinstance(row_values, list)

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is None, 'already used time_stamp: %r'%(try_time_stamp)
        time_stamp = try_time_stamp

        assert len(row_values) == self.__phase.columns.size

        # create a row with object data type; users row_values data define data type
        row = pandas.DataFrame( index=[time_stamp], 
                                columns=list( self.__phase.columns ), dtype=object )

        for (col,v) in zip(row.columns, row_values):
            row.loc[time_stamp,col] = v

        frames = [self.__phase, row]
        self.__phase = pandas.concat(frames)
        return

    def GetRow(self, try_time_stamp=None):
        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)
        return list(self.__phase.loc[time_stamp, :])

    def GetColumn(self, actor):
        assert isinstance(actor, str)
        assert actor in self.__phase.columns, 'actor %r not in %r'% \
                   (actor,self.__phase.columns)
        return list(self.__phase.loc[:, actor])

    def ScaleRow(self, try_time_stamp, value):
        assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float) 
        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)
        assert isinstance(value, int) or isinstance(value, float) 
        self.__phase.loc[time_stamp, :] *= value
        return

    # set species and quantities of history to a given value (default to zero value)
    # all time stamps are preserved
    def ClearHistory(self, value=0.0):
        assert isinstance(value, int) or isinstance(value, float) 
        self.__phase.loc[:, :] = value
        return

    # set species and quantities of history to a given value (default to zero value)
    # only one time stamp is preserved (default to last time stamp)
    def ResetHistory(self, try_time_stamp=None, value=None):
        if value is not None:
           assert isinstance(value, int) or isinstance(value, float) or \
                  isinstance(value, npy.ndarray)

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float) 

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        values = self.GetRow(time_stamp)  # save values

        columns = list(self.__phase.columns)
        assert len(columns) == len(values), 'FATAL: oops internal error.'

        self.__phase = pandas.DataFrame( index=[time_stamp], columns=columns )
        self.__phase.fillna( 0.0, inplace=True )

        if value is None:
            for v in values:
                idx = values.index(v)
                self.__phase.loc[time_stamp, columns[idx]] = v  # restore values
        else:
            self.__phase.loc[time_stamp, :] = value   # set user-given value

        return

    def GetValue(self, actor, try_time_stamp=None):

        assert isinstance(actor, str)
        assert actor in self.__phase.columns, 'actor %r not in %r'% \
                   (actor,self.__phase.columns)

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float) 

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

        return self.__phase.loc[time_stamp, actor]

# old def SetValue(self, time_stamp, actor, value):
# new
    def SetValue(self, actor, value, try_time_stamp=None):

        assert isinstance(actor, str)
        assert actor in self.__phase.columns

#        assert isinstance(value, int) or isinstance(value, float) or \
#               isinstance(value, npy.ndarray)

        if try_time_stamp is not None:
           assert isinstance(try_time_stamp, int) or isinstance(try_time_stamp, float) 

        time_stamp = self.__get_time_stamp( try_time_stamp )
        assert time_stamp is not None, 'missing try_time_stamp: %r'%(try_time_stamp)

#        print('*value =', value)
#        print('*type(value) =', type(value))
#        print('*time_stamp =', time_stamp)
#        print('*actor =', actor)
#        print('*column values =', self.__phase[actor])
#        print('*df =', self.__phase)
#        print('*df.dtypes =', self.__phase.dtypes)
#        print('*df.shape =', self.__phase.shape)
#        print('')
#        if isinstance(value,npy.ndarray):
#           self.__phase[actor] = self.__phase.astype({actor:type(value)})
#        print('*df.dtypes =', self.__phase.dtypes)
#        print('')
        self.__phase.loc[time_stamp, actor] = value
        return

    def WriteHTML(self, fileName):
        assert isinstance(fileName, str)
        tmp = pandas.DataFrame(self.__phase)
        columnNames = tmp.columns
        speciesNames = [specie.name for specie in self.__species]
        quantityNames = [quantity.name for quantity in self.__quantities]
        for col in columnNames:
            if col in speciesNames:
                idx = speciesNames.index(col)
                specie = self.__species[idx]
                tmp.rename(columns={col: specie.formulaName}, inplace=True)
            elif col in quantityNames:
                idx = quantityNames.index(col)
                quant = self.__quantities[idx]
                tmp.rename( columns={ col: col + '[' + quant.unit + ']'},
                            inplace=True )
            else:
                assert False, 'oops fatal.'
        tmp.to_html(fileName)

#*********************************************************************************
# Private helper functions (internal use: __)

    def __get_time_stamp(self, try_time_stamp):
        """
        Helper method for finding the closest time stamp in the phase history.  
        The pandas Index container used for storing float data type time stamps
        will return the nearest time stamp up to a tolerance.
        """

        if try_time_stamp is None:
           return self.__phase.index[-1]
        else:
           tol = 1.0e-4
           try:
             loc = self.__phase.index.get_loc( try_time_stamp, method='nearest', 
                                               tolerance=tol )
           except KeyError:
             return None
           else:
             return  self.__phase.index[loc]

# *******************************************************************************
# Printing of data members
# def __str__( self ):
        s = '\n\t **Phase()**: \n\t *quantities*: %s\n\t *species*: %s\n\t *history* #time_stamp=%s\n\t *history end* @%s\n%s'
        return s % (self.__quantities, self.__species, len(self.__phase.index),
                    self.__phase.index[-1], self.__phase.loc[self.__phase.index[-1], :])
#

    def __repr__(self):
        s = '\n\t **Phase()**: \n\t *quantities*: %s\n\t *species*: %s\n\t *history* #time_stamp=%s\n\t *history end* @%s\n%s'
        return s % (self.__quantities, self.__species, len(self.__phase.index),
                    self.__phase.index[-1], self.__phase.loc[self.__phase.index[-1], :])
# *******************************************************************************



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda

Nuclides container.
The purpose of the this container is to store and query a table of nuclides.
Typically the table is filled in with data from an ORIGEN calculation or
some other fission/transmutation code.

VFdALib support classes

Sat Jun 27 14:46:49 EDT 2015
"""

# *******************************************************************************
import os
import sys
import io
import time
import datetime
import math
import random
import pandas

from cortix.support.periodictable import ELEMENTS
from cortix.support.periodictable import SERIES
# *******************************************************************************

# *******************************************************************************


class Nuclides():

    def __init__(self,
                 propertyDensities=pandas.DataFrame()
                 ):

        assert isinstance(
            propertyDensities, type(
                pandas.DataFrame())), 'fatal.'

        self.attributeNames = \
            ['nuclides',
             'isotopes',
             'massDens',
             'massCC',
             'radioactivityDens',
             'thermalDens',
             'heatDens',
             'gammaDens']

        self.chemicalElementSeries = \
            ['alkali metals', 'alkali earth metals', 'lanthanides', 'actinides',
             'transition metals', 'noble gases', 'metalloids', 'fission products', 'nonmetals',
             'oxide fission products', 'halogens', 'minor actinides', 'volatile fission products', 'poor metals']

        self.propertyDensities = propertyDensities

        return

# *******************************************************************************
# Setters and Getters methods
# -------------------------------------------------------------------------------
# These are passing arguments by value effectively. Because the python objects
# passed into/out of the function are immutable.

    def GetAttribute(self, name, symbol=None, series=None):
        return self.__GetAttribute(name, symbol, series)


# *********************************************************************************

# Get property either overall or on a nuclide basis

# ---------------------------------------------------------------------------------

    def __GetAttribute(self, attributeName, symbol=None, series=None):

        assert attributeName in self.attributeNames, ' attribute name: %r; options: %r; fail.' % (
            attributeName, self.attributeNames)

        if symbol is not None:
            assert series is None, 'fail.'
        if series is not None:
            assert symbol is None, 'fail.'

        if attributeName == 'isotopes':
            assert symbol is not None, 'need an element symbol.'

# .................................................................................
# isotopes python list

        if attributeName == 'isotopes':

            nuclidesNames = self.propertyDensities.index
            isotopes = [x for x in nuclidesNames if x.split(
                '-')[0].strip() == symbol]
            return isotopes

# .................................................................................
# nuclides python list

        if attributeName == 'nuclides':

            if series is not None:
                # CREATE A HELPER FUNCTION FOR THIS; NOTE THIS IS USED BELOW
                # TOO!!!
                nuclidesNames = self.propertyDensities.index

                seriesNameMap = {
                    'alkali metals': 'Alkali metals',
                    'alkali earth metals': 'Alkaline earth metals',
                    'lanthanides': 'Lanthanides',
                    'actinides': 'Actinides',
                    'transition metals': 'Transition metals',
                    'noble gases': 'Noble gases',
                    'metalloids': 'Metalloids',
                    'fission products': 'fission products',
                    'nonmetals': 'Nonmetals',
                    'oxide fission products': 'oxide fission products',
                    'halogens': 'Halogens',
                    'minor actinides': 'minor actnides',
                    'volatile fission products': 'volatile fission products',
                    'poor metals': 'Poor metals'}
                #
                if series == 'fission products':
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] != seriesNameMap['actinides']]
                #
                elif series == 'oxide fission products':
                    collec = [seriesNameMap['actinides'],
                              seriesNameMap['halogens'],
                              seriesNameMap['noble gases']
                              ]
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] not in collec]
                    collec = ['C', 'N', 'O', 'H']
                    nuclides = [x for x in nuclides if x.split(
                        '-')[0].strip() not in collec]
                #
                elif series == 'volatile fission products':
                    collec = [seriesNameMap['actinides'],
                              seriesNameMap['alkali metals'],
                              seriesNameMap['alkali earth metals'],
                              seriesNameMap['lanthanides'],
                              seriesNameMap['metalloids'],
                              seriesNameMap['transition metals'],
                              seriesNameMap['poor metals']
                              ]
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] not in collec]
                    collec = ['C', 'O']
                    nuclides = [x for x in nuclides if x.split(
                        '-')[0].strip() not in collec]
                #
                elif series == 'minor actinides':
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] == seriesNameMap['actinides']]
                    collec = ['U', 'Pu']
                    nuclides = [x for x in nuclides if x.split(
                        '-')[0].strip() not in collec]
                #
                else:
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] == seriesNameMap[series]]

                return nuclides

            if series is None:

                nuclidesNames = self.propertyDensities.index
                return list(nuclidesNames)

# .................................................................................
# mass or mass concentration

        if attributeName == 'massCC':
            colName = 'Mass CC [g/cc]'

# .................................................................................
# radioactivity

        if attributeName == 'radioactivityDens':
            colName = 'Radioactivity Dens. [Ci/cc]'

# .................................................................................
# thermal

        if attributeName == 'thermalDens' or attributeName == 'heatDens':
            colName = 'Thermal Dens. [W/cc]'

# .................................................................................
# gamma

        if attributeName == 'gammaDens':
            colName = 'Gamma Dens. [W/cc]'

# .................................................................................
##########################################################################
# .................................................................................

# .................................................................................
# all nuclide content added

        if symbol is None and series is None:

            density = 0.0
            density = self.propertyDensities[colName].sum()
            return float(density)  # avoid numpy.float64 type

# .................................................................................
# get chemical element series

        if series is not None:

            density = 0.0

            assert series in self.chemicalElementSeries, 'series: %r; fail.' % (
                series)

            seriesNameMap = {
                'alkali metals': 'Alkali metals',
                'alkali earth metals': 'Alkaline earth metals',
                'lanthanides': 'Lanthanides',
                'actinides': 'Actinides',
                'transition metals': 'Transition metals',
                'noble gases': 'Noble gases',
                'metalloids': 'Metalloids',
                'fission products': 'fission products',
                'nonmetals': 'Nonmetals',
                'oxide fission products': 'oxide fission products',
                'halogens': 'Halogens',
                'minor actinides': 'minor actnides',
                'volatile fission products': 'volatile fission products',
                'poor metals': 'Poor metals'}

            if series in self.chemicalElementSeries:

                nuclidesNames = self.propertyDensities.index
                #
                if series == 'fission products':
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] != seriesNameMap['actinides']]
                #
                elif series == 'oxide fission products':
                    collec = [seriesNameMap['actinides'],
                              seriesNameMap['halogens'],
                              seriesNameMap['noble gases']
                              ]
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] not in collec]
                    collec = ['C', 'N', 'O', 'H']
                    nuclides = [x for x in nuclides if x.split(
                        '-')[0].strip() not in collec]
                #
                elif series == 'volatile fission products':
                    collec = [seriesNameMap['actinides'],
                              seriesNameMap['alkali metals'],
                              seriesNameMap['alkali earth metals'],
                              seriesNameMap['lanthanides'],
                              seriesNameMap['metalloids'],
                              seriesNameMap['transition metals'],
                              seriesNameMap['poor metals']
                              ]
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] not in collec]
                    collec = ['C', 'O']
                    nuclides = [x for x in nuclides if x.split(
                        '-')[0].strip() not in collec]
                #
                elif series == 'minor actinides':
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] == seriesNameMap['actinides']]
                    collec = ['U', 'Pu']
                    nuclides = [x for x in nuclides if x.split(
                        '-')[0].strip() not in collec]
                #
                else:
                    nuclides = [x for x in nuclidesNames if SERIES[ELEMENTS[x.split(
                        '-')[0].strip()].series] == seriesNameMap[series]]

#       print('fission products ',nuclides)

                for nuclide in nuclides:
                    density += self.propertyDensities.loc[nuclide, colName]

            return float(density)  # avoid numpy.float64 type

# .................................................................................
# get specific nuclide (either the isotopes of the nuclide or the specific
# isotope) property

        if symbol is not None:

            density = 0.0

        # single isotope
            if len(symbol.split('-')) == 2:
                density = self.propertyDensities.loc[symbol, colName]

        # many isotopes
            else:
                nuclidesNames = self.propertyDensities.index
#    print(self.propertyDensities)
                isotopes = [
                    x for x in nuclidesNames if x.split('-')[0].strip() == symbol]
#    print(isotopes)
                for isotope in isotopes:
                    density += self.propertyDensities.loc[isotope, colName]

            return float(density)  # avoid numpy.float64.type

# *********************************************************************************

# *******************************************************************************
# Printing of data members
# def __str__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formulaName, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)
#
# def __repr__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formulaName, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)
# *******************************************************************************



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
"""
Author: Valmor de Almeida dealmeidav@ornl.gov; vfda

Fuel segment

VFdALib support classes

Sat Jun 27 14:46:49 EDT 2015
"""

# *******************************************************************************
import os
import sys
import io
import time
import datetime
import math
import random

from cortix.support.fuel_segment import FuelSegment
# *******************************************************************************

# *******************************************************************************


class FuelSegmentsGroups():

    def __init__(self,
                 key=None,
                 fuelSegments=None
                 ):

        # *******************************************************************************
        # constructor
        # FuelSegmentsGroups simply encapsulates a dictionary of a list of
        # FuelSegment objects. The key is typically a time stamp.
        # A FuelSegment object has two data members, a pandas Series for geometry spec
        # and a pandas DataFrame for property density.

        # This is the central member data
        self.groups = dict()

        assert key not in self.groups.keys()

        if fuelSegments is not None:
            assert isinstance(fuelSegments, list)
            assert isinstance(fuelSegments[-1], FuelSegment)
            self.groups[key] = fuelSegments
        else:
            self.groups[key] = list()

        return

# *******************************************************************************
# Setters and Getters methods
# -------------------------------------------------------------------------------
# These are passing arguments by value effectively. Because the python objects
# passed into/out of the function are immutable.

    def HasGroup(self, key):

        return key in self.groups.keys()

    def AddGroup(self, key, fuelSegments=None):

        self.__AddGroup(key, fuelSegments)

    def GetAttribute(self, groupKey=None, attributeName=None,
                     nuclideSymbol=None, nuclideSeries=None):

        return self.__GetGroupAttribute(
            groupKey, attributeName, nuclideSymbol, nuclideSeries)

    def GetFuelSegments(self, groupKey=None):

        return self.__GetFuelSegments(groupKey)

    def RemoveFuelSegment(self, groupKey, fuelSegment):

        return self.__RemoveFuelSegment(groupKey, fuelSegment)


# *********************************************************************************


# If an attibute is not found, Return 0 even if a groupKey is not found
# Don't change this behavior; it will break user's code.
# ---------------------------------------------------------------------------------

    def __GetGroupAttribute(self, groupKey=None,
                            attributeName=None, symbol=None, series=None):

        assert attributeName is not None, 'fatal.'

        attribute = None

# Either cumulative or average density property for all fuel segments for *all* groups
# BE VERY CAREFUL HERE: groups with no segments will reduce the average
# density value
        if groupKey is None:

            attribute = 0

            for (key, fuelSegments) in self.groups.items():
                assert isinstance(fuelSegments, list), 'fail.'
                if len(fuelSegments) == 0:
                    continue  # this will reduce the average value
                groupAttribute = 0
                for fuelSegment in fuelSegments:
                    groupAttribute += fuelSegment.get_attribute(
                        attributeName, symbol, series)

                if attributeName[-4:] == 'Dens' or attributeName[-2:] == 'CC':
                    groupAttribute /= len(fuelSegments)
                attribute += groupAttribute

            if attribute != 0 and \
               (attributeName[-4:] == 'Dens' or attributeName[-2:] == 'CC'):
                attribute /= len(self.groups)

#     if attributeName[-4:] == 'Dens' or attributeName[-2:] == 'CC':
#        print('HELLO density '+attributeName+' ', attribute)

# Get average property in all fuel segments within a groupKey
        else:

            if groupKey not in self.groups.keys():
                return 0

            fuelSegments = self.groups[groupKey]

            if len(fuelSegments) is 0:
                return 0

            attribute = 0

            for fuelSegment in fuelSegments:
                attribute += fuelSegment.get_attribute(
                    attributeName, symbol, series)

            if attribute != 0 and \
               (attributeName[-4:] == 'Dens' or attributeName[-2:] == 'CC'):
                attribute /= len(fuelSegments)

        return attribute

# *********************************************************************************

# Return the fuel segments of a given group (if the groupKey is given), otherwise
# returns an ordered list of pairs of all segments in all groups and their keys.
# That is, [ (timeStamp,fuelSegment), (timeStamp,fuelSegment), ... ]

# *ALWAYS* return the list of segments held by the group when a groupKey is given.

# If the groupKey does not exist return an empty list()
# ---------------------------------------------------------------------------------
    def __GetFuelSegments(self, groupKey=None):

        if groupKey is None:  # return an ordered list of all fuelSegments

            tmp = list()
            timeStamp = list()
            for (key, fuelSegments) in self.groups.items():
                if fuelSegments is None:
                    continue
                tmp += fuelSegments
                # all fuel segments in the group have
                timeStamp += [key for i in fuelSegments]
                # the same time stamp

            # sort fuelSegments in order of their keys
            data = zip(timeStamp, tmp)  # this is a list of pairs
            sorted_data = sorted(
                data, key=lambda entry: entry[0], reverse=False)
#    tmp = [ y for (x,y) in sorted_data ]  # oldest first
            return sorted_data

        else:

            if groupKey not in self.groups.keys():
                return list()

            return self.groups[groupKey]

# *********************************************************************************

# Make fuelSegments *ALWAYS* a list (could be empty)
# If group key exists, add to group otherwise create a group
# If fuelSegments are not given, add an empty list to a group if it exists
# A group will *ALWAYS* have a fuelSegments list.

# ---------------------------------------------------------------------------------
    def __AddGroup(self, groupKey, fuelSegments=None):

        if fuelSegments is None:
            fuelSegments = list()
        else:
            assert isinstance(fuelSegments, list), 'fail.'

        if groupKey in self.groups.keys():
            self.groups[groupKey] += fuelSegments
        else:
            self.groups[groupKey] = fuelSegments

# *********************************************************************************

# ---------------------------------------------------------------------------------
    def __RemoveFuelSegment(self, groupKey, fuelSegment_remove):

        assert groupKey in self.groups.keys(), 'fail.'

        fuelSegments = self.groups[groupKey]
        nSegments = len(fuelSegments)

        for fuelSegment in fuelSegments:
            if fuelSegment.get_attribute(
                    'segmentId') == fuelSegment_remove.get_attribute('segmentId'):
                fuelSegments.remove(fuelSegment)

        assert len(self.groups[groupKey]) == nSegments - 1, 'fatal.'

# *********************************************************************************

# *******************************************************************************
# Printing of data members
# def __str__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formula, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)
#
# def __repr__( self ):
#     s = ' %5s %5s %5s '+' molar mass: %6s '+' molar cc: %6s '+' mass cc: %6s '+' flag: %s '+'# atoms: %s'+' atoms: %s\n'
#     return s % (self.name, self.formula, self.phase, self.molarMass, self.molarCC, self.massCC, self.flag, self.nAtoms, self.atoms)
# *******************************************************************************



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
PyPlot module.

Author: Valmor F. de Almeida dealmeidav@ornl.gov; vfda
Tue Jun 24 01:03:45 EDT 2014
'''
#*********************************************************************************
import os
import sys
import io
import time
import datetime
import logging
import xml.etree.ElementTree as ElementTree
import numpy as npy
import matplotlib
matplotlib.use('Agg', warn=False)
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.ticker import MultipleLocator
import pickle

import xml.etree.ElementTree as ElementTree
from threading import Lock

from cortix.src.utils.xmltree import XMLTree

from cortix.modulib.pyplot.time_sequence import TimeSequence
# *********************************************************************************

class PyPlot():

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self,
                 slot_id,
                 input_full_path_file_name,
                 manifest_full_path_file_name,
                 work_dir,
                 ports = list(),
                 cortix_start_time = 0.0,
                 cortix_final_time = 0.0,
                 cortix_time_step = 0.0,
                 cortix_time_unit  = None
                 ):

        # Sanity test
        assert isinstance( slot_id, int), '-> slot_id type %r is invalid.' % \
                           type(slot_id)
        assert isinstance( ports, list), '-> ports type %r is invalid.' % type(ports)
        assert ports is not None, 'fatal.'
        assert isinstance( cortix_start_time, float), '-> time type %r is invalid.' % \
                           type(cortix_start_time)
        assert isinstance( cortix_final_time, float), '-> time type %r is invalid.' % \
                           type(cortix_final_time)
        assert isinstance( cortix_time_step, float), '-> time step type %r is invalid.' % \
                           type(cortix_time_step)

        assert isinstance( cortix_time_unit, str), '-> time unit type %r is invalid.' % \
                           type(cortix_time_unit)

        # Logger
        self.__log = logging.getLogger('launcher-pyplot_' + str(slot_id) +
                                       '.cortix_driver.pyplot')
        self.__log.info('initializing an object of PyPlot()')

        # Read the manisfest
        self.__read_manifest( manifest_full_path_file_name )
        self.__log.info(self.__port_diagram)

        # Member data
        self.__slot_id = slot_id
        self.__ports   = ports

        self.__cortix_start_time = cortix_start_time
        self.__cortix_final_time = cortix_final_time
        self.__cortix_time_unit  = cortix_time_unit

        n_time_steps = ( cortix_final_time - cortix_start_time ) / cortix_time_step
        n_of_times_to_plot = 10 # during the entire run
        # how often to plot
        self.__plot_interval = round( n_time_steps / n_of_times_to_plot )
        # the width of the window to plot
        self.__plot_slide_window_interval = 3 * self.__plot_interval

        s = 'Plot interval = ' + str(self.__plot_interval) + ' [' + self.__cortix_time_unit + ']'
        self.__log.info(s)
        s = 'Plot slide window interval = ' + str(self.__plot_slide_window_interval) + ' [' + self.__cortix_time_unit + ']'
        self.__log.info(s)

        # This holds all time sequences, potentially, one per use port. One time sequence
        # has all the data for one port connection.
        # Time sequences are xml formatted data per Cortix convention.
        self.__time_sequences_tmp = list()  # temporary storage

        # Tables in xml format per Cortix convention
        # [(time,timeUnit)] = [column,column,...]
        self.__time_tables_data = dict(list())

# .................................................................................
# Input ports (if any)

        fin = open(input_full_path_file_name, 'r')
        input_data_full_path_file_names = list()
        for line in fin:
            input_data_full_path_file_names.append( line.strip() )
        fin.close()

        if len(input_data_full_path_file_names) == 0:
            return

        found = False
        for (portName, portType, port_file) in ports:
            if portName == 'input':  # this is the use port connected to the input port
                s = 'cp -f ' + input_data_full_path_file_names[0] + ' ' + port_file
                os.system(s)
                self.__log.debug(s)
                found = True

        if found is True:
            s = 'found input file.'
            self.__log.info(s)
        else:
            s = 'did not find input file.'
            self.__log.warn(s)

        return

#*********************************************************************************
# Public member functions
#*********************************************************************************

    def call_ports(self, cortix_time=0.0):
        '''
        Transfer data at cortix_time
        '''

        if (cortix_time % self.__plot_interval <= 1e-2 and \
            cortix_time < self.__cortix_final_time)        \
           or cortix_time >= self.__cortix_final_time:

            # use ports in PyPlot have infinite multiplicity (implement
            # multiplicity later)
            assert len(self.__time_sequences_tmp) == 0

            for port in self.__ports:
                (portName, portType, thisPortFile) = port

                if portType == 'use':
                    assert portName == 'time-sequence' or portName == 'time-tables' or \
                        portName == 'input'

                    self.__use_data( usePortName = portName,
                                     usePortFile = thisPortFile,
                                     at_time = cortix_time )

        return

    def execute(self, cortix_time=0.0, time_step=0.0):

        s = 'execute(): cortix time [min] = ' + str(round(cortix_time, 3))
        self.__log.debug(s)

        self.__plot_data(cortix_time, time_step)

        return

#*********************************************************************************
# Private helper functions (internal use: __)
#*********************************************************************************

    def __use_data(self, usePortName=None, usePortFile=None, at_time=0.0):
        '''
        This operates on a given use port;
        '''

        # Access the port file
        port_file = self.__get_port_file( usePortName = usePortName,
                                          usePortFile = usePortFile  )

# Get data from port files
        if usePortName == 'time-sequence' or usePortName == 'input' \
           and port_file is not None:

            self.__get_time_sequence( port_file, at_time )

        if usePortName == 'time-tables' and port_file is not None:
            _GetTimeTables(self, port_file, at_time)

        return

    def __provide_data(self, providePortName=None, at_time=0.0):
        '''
        Nothing planned to provide at this time; but could change
        '''

        return

    def __get_port_file(self, usePortName=None,
                        usePortFile=None, providePortName=None):

        port_file = None

        # ..........
        # Use ports
        # ..........
        if usePortName is not None:

            assert providePortName is None

            for port in self.__ports:
                (portName, portType, thisPortFile) = port
                if portName == usePortName and portType == 'use' and thisPortFile == usePortFile:
                    port_file = thisPortFile

            if port_file is None:
                return None

            max_n_trials = 50
            n_trials = 0
            while os.path.isfile(port_file) is False and n_trials <= max_n_trials:
                n_trials += 1
                time.sleep(0.1)

            if n_trials > max_n_trials:
                s = '__get_port_file(): waited ' + str(n_trials) + \
                    ' trials for port: ' + port_file
                self.__log.warn(s)

            assert os.path.isfile(port_file) is True, 'port_file %r not available; stop.' % port_file

        # ..............
        # Provide ports
        # ..............
        if providePortName is not None:

            assert usePortName is None

            for port in self.__ports:
                (portName, portType, thisPortFile) = port
                if portName == providePortName and portType == 'provide':
                    port_file = thisPortFile

        return port_file

    def __plot_data(self, cortix_time=0.0, time_step=0.0):

        if cortix_time % self.__plot_interval <= 1e-2 and \
           cortix_time < self.__cortix_final_time:

            s = '__plot_data(): cortix time [min] = ' + str(cortix_time)
            self.__log.debug(s)

            from_time = max(0.0, cortix_time - self.__plot_slide_window_interval)
            to_time = cortix_time

            s = '__plot_data(): from_time [min] = ' + str(from_time)
            self.__log.debug(s)
            s = '__plot_data(): to_time [min] = ' + str(to_time)
            self.__log.debug(s)

            # plot with slide window history
            self.__plot_time_seq_dashboard( from_time, to_time )

            # plot with slide window history
            self.__plot_time_tables( from_time, to_time )

        elif cortix_time >= self.__cortix_final_time:

            s = '__plot_data(): cortix time [min] = ' + str(cortix_time)
            self.__log.debug(s)

            self.__plot_time_seq_dashboard( self.__cortix_start_time, self.__cortix_final_time)  # plot all history

            self.__plot_time_tables( self.__cortix_start_time, self.__cortix_final_time )  # plot all history

        return

    def __get_time_sequence(self, port_file, at_time):
        '''
        This uses a use port_file which is guaranteed to exist at this point
        '''

        s = '__get_time_sequence(): will get data in port_file: ' + port_file
        self.__log.debug(s)

        if at_time >= self.__cortix_final_time:
            initialTime = self.__cortix_start_time
        else:
            initialTime = max(self.__cortix_start_time, at_time - self.__plot_slide_window_interval)

        timeSequence = TimeSequence( port_file, 'xml', initialTime, at_time,
                                     self.__cortix_time_unit, self.__log )

        self.__time_sequences_tmp.append(timeSequence)

        s = '__get_time_sequence(): loaded ' + timeSequence.get_name()
        self.__log.debug(s)

        return

    def __get_time_tables(self, port_file, at_time):
        '''
        This uses a use port_file which is guaranteed at this point
        '''

        s = '__get_time_tables(): will check file: ' + port_file
        self.__log.debug(s)

        found = False

        while found is False:

            s = '__get_time_tables(): checking for value at ' + str(at_time)
            self.__log.debug(s)

#    tree = ElementTree.parse(port_file)

            try:
                mutex = Lock()
                mutex.acquire()
                tree = ElementTree.parse(port_file)
            except ElementTree.ParseError as error:
                mutex.release()
                s = '__get_time_tables(): ' + port_file + ' unavailable. Error code: ' + \
                    str(error.code) + ' File position: ' + \
                    str(error.position) + '. Retrying...'
                self.__log.debug(s)
                time.sleep(0.1)
                continue

            mutex.release()
            rootNode = tree.getroot()
            assert rootNode.tag == 'time-tables', 'invalid format.'

            timeNodes = rootNode.findall('timeStamp')

            for timeNode in timeNodes:

                timeStamp = float(timeNode.get('value').strip())

                if timeStamp == at_time:

                    found = True

                    timeUnit = timeNode.get('unit').strip()

                    self.__time_tables_data[(timeStamp, timeUnit)] = list()

                    columns = timeNode.findall('column')

                    data = list()

                    for col in columns:
                        data.append(col)

                    self.__time_tables_data[(timeStamp, timeUnit)] = data

                    s = '__get_time_tables(): added ' + str(len(data)) + ' columns of data'
                    self.__log.debug(s)

                # end of if timeStamp == at_time:

            # end of for timeNode in timeNodes:

            if found is False:
                time.sleep(0.1)

        # while found is False:

        return

    def __plot_time_seq_dashboard( self, initialTime=0.0, finalTime=0.0 ):
        '''
        All time sequences on hold will be plotted here and the temporary storage
        of these time sequences will be cleared at the end.
        '''

        nRows = 3
        nCols = 1

        nSequences = len(self.__time_sequences_tmp)
        if nSequences == 0:
            return

        s = '__plot_time_seq_dashboard(): from: ' + str(initialTime) + \
            ' to ' + str(finalTime)
        self.__log.debug(s)

        s = '__plot_time_seq_dashboard(): # of sequences: ' + str(nSequences)
        self.__log.debug(s)

        nVar = 0
        for seq in self.__time_sequences_tmp:
            nVar += seq.GetNVariables()

        s = '__plot_time_seq_dashboard(): # of variables: ' + str(nVar)
        self.__log.debug(s)

    # multiple sequences are stored coming from various sources
    # collect all variables in a list for mapping on the dashboards
        variablesData = list()
        for seq in self.__time_sequences_tmp:
            for (spec, values) in seq.GetVariables().items():
                variablesData.append((spec, values))

        assert len(variablesData) == nVar, 'len(variablesData) = %r; nVar = %r' % (
            len(variablesData), nVar)

        today = datetime.datetime.today().strftime("%d%b%y %H:%M:%S")

        figNum = None

        # loop over variables and assign to the dashboards
        iDash = 0
        for iVar in range(nVar):

            if iVar % (
                    nRows * nCols) == 0:  # if a multiple of nRows*nCols start a new dashboard

                if iVar != 0:  # flush any current figure
                    figName = 'pyplot_' + \
                        str(self.__slot_id) + '-timeseq-dashboard-' + \
                        str(iDash).zfill(2)
                    fig.savefig(figName+'.png', dpi=200, fomat='png')
                    plt.close(figNum)

                    pickle.dump( fig, open(figName+'.pickle','wb') )

                    s = '__plot_time_seq_dashboard(): created plot: ' + figName
                    self.__log.debug(s)

                    iDash += 1
                # end of: if iVar != 0: # flush any current figure

                figNum = str(self.__slot_id) + '.' + str(iDash)
                fig = plt.figure(num=figNum)

                gs = gridspec.GridSpec(nRows, nCols)
#                gs.update(left=0.08, right=0.98, wspace=0.4, hspace=0.4)
                gs.update(left=0.09, right=0.98, wspace=0.4, hspace=0.5)

                axlst = list()

                nPlotsNeeded = nVar - iVar
                count = 0
                for i in range(nRows):
                    for j in range(nCols):
                        axlst.append(fig.add_subplot(gs[i, j]))
                        count += 1
                        if count == nPlotsNeeded:
                            break
                    if count == nPlotsNeeded:
                        break

                axes = npy.array(axlst)

                text = today + ': Cortix.PyPlot_' + \
                    str(self.__slot_id) + ': Time-Sequence Dashboard'
                fig.text(.5, .95, text, horizontalalignment='center', fontsize=14)

                axs = axes.flat

                axId = 0

            # end of: if iVar % nRows*nCols == 0: # if a multiple of nRows*nCols
            # start a new dashboard

            (spec, val) = variablesData[iVar]

            ax = axs[axId]
            axId += 1

            varName = spec[0]
            varUnit = spec[1]

            if varUnit == 'gram':
                varUnit = 'g'
            if varUnit == 'gram/min':
                varUnit = 'g/min'
            if varUnit == 'gram/s':
                varUnit = 'g/s'
            if varUnit == 'gram/m3':
                varUnit = 'g/m3'
            if varUnit == 'gram/L':
                varUnit = 'g/L'
            if varUnit == 'sec':
                varUnit = 's'

            timeUnit = spec[2]
            varLegend = spec[3]
            varScale = spec[4]
            assert varScale == 'log' or varScale == 'linear' or varScale == 'log-linear' \
                or varScale == 'linear-log' or varScale == 'linear-linear' or \
                varScale == 'log-log'

            if timeUnit == 'minute':
                timeUnit = 'min'

            data = npy.array(val)  # convert to numpy ndarray

#      assert len(data.shape) == 2, 'not a 2-column shape: %r in var %r of %r; stop.' % (data.shape,varName,varLegend)
            if len(data.shape) != 2:
                s = '__plot_time_seq_dashboard(): bad data; variable: ' + varName + ' unit: ' + \
                    varUnit + ' legend: ' + varLegend + \
                    ' shape: ' + str(data.shape) + ' skipping...'
                self.__log.warn(s)
                continue  # simply skip bad data and log

            x = data[:, 0]

            if (varScale == 'linear' or varScale == 'linear-linear' or \
                varScale == 'linear-log') and x.max() >= 120.0:
                x /= 60.0
                if timeUnit == 'min':
                    timeUnit = 'h'

            y = data[:, 1]

            if y.max() >= 1e3 and varScale != 'linear-log' and \
                    varScale != 'log-log' and varScale != 'log':
                y /= 1e3
                if varUnit == 'gram' or varUnit == 'g':
                    varUnit = 'kg'
                if varUnit == 'L':
                    varUnit = 'kL'
                if varUnit == 'cc':
                    varUnit = 'L'
                if varUnit == 'Ci':
                    varUnit = 'kCi'
                if varUnit == 'W':
                    varUnit = 'kW'
                if varUnit == 'gram/min' or varUnit == 'g/min':
                    varUnit = 'kg/min'
                if varUnit == 'gram/s' or varUnit == 'g/s':
                    varUnit = 'kg/s'
                if varUnit == 'gram/m3' or varUnit == 'g/m3':
                    varUnit = 'kg/m3'
                if varUnit == 'gram/L' or varUnit == 'g/L':
                    varUnit = 'kg/L'
                if varUnit == 'W/L':
                    varUnit = 'kW/L'
                if varUnit == 'Ci/L':
                    varUnit = 'kCi/L'
                if varUnit == '':
                    varUnit = 'x1e3'
                if varUnit == 'L/min':
                    varUnit = 'kL/min'
                if varUnit == 'Pa':
                    varUnit = 'kPa'
                if varUnit == 's':
                    varUnit = 'ks'
                if varUnit == 'm':
                    varUnit = 'km'

            if y.max() < 1e-6 and varScale != 'linear-log' and \
                    varScale != 'log-log' and varScale != 'log':
                y *= 1e9
                if varUnit == 'gram' or varUnit == 'g':
                    varUnit = 'ng'
                if varUnit == 'cc':
                    varUnit = 'n-cc'
                if varUnit == 'L':
                    varUnit = 'nL'
                if varUnit == 'W':
                    varUnit = 'nW'
                if varUnit == 'Ci':
                    varUnit = 'nCi'
                if varUnit == 'gram/min' or varUnit == 'g/min':
                    varUnit = 'ng/min'
                if varUnit == 'gram/s' or varUnit == 'g/s':
                    varUnit = 'ng/s'
                if varUnit == 'gram/m3' or varUnit == 'g/m3':
                    varUnit = 'ng/m3'
                if varUnit == 'gram/L' or varUnit == 'g/L':
                    varUnit = 'ng/L'
                if varUnit == 'W/L':
                    varUnit = 'nW/L'
                if varUnit == 'Ci/L':
                    varUnit = 'nCi/L'
                if varUnit == 'L/min':
                    varUnit = 'nL/min'
                if varUnit == 'Pa':
                    varUnit = 'nPa'
                if varUnit == 's':
                    varUnit = 'ns'

            if (y.max() >= 1e-6 and y.max() < 1e-3) and varScale != 'linear-log' and \
                    varScale != 'log-log' and varScale != 'log':
                y *= 1e6
                if varUnit == 'gram' or varUnit == 'g':
                    varUnit = 'ug'
                if varUnit == 'cc':
                    varUnit = 'u-cc'
                if varUnit == 'L':
                    varUnit = 'uL'
                if varUnit == 'W':
                    varUnit = 'uW'
                if varUnit == 'Ci':
                    varUnit = 'uCi'
                if varUnit == 'gram/min' or varUnit == 'g/min':
                    varUnit = 'ug/min'
                if varUnit == 'gram/s' or varUnit == 'g/s':
                    varUnit = 'ug/s'
                if varUnit == 'gram/m3' or varUnit == 'g/m3':
                    varUnit = 'ug/m3'
                if varUnit == 'gram/L' or varUnit == 'g/L':
                    varUnit = 'ug/L'
                if varUnit == 'W/L':
                    varUnit = 'uW/L'
                if varUnit == 'Ci/L':
                    varUnit = 'uCi/L'
                if varUnit == 'L/min':
                    varUnit = 'uL/min'
                if varUnit == 'Pa':
                    varUnit = 'uPa'
                if varUnit == 's':
                    varUnit = 'us'

            if (y.max() >= 1e-3 and y.max() < 1e-1) and varScale != 'linear-log' and \
                    varScale != 'log-log' and varScale != 'log':
                y *= 1e3
                if varUnit == 'gram' or varUnit == 'g':
                    varUnit = 'mg'
                if varUnit == 'cc':
                    varUnit = 'm-cc'
                if varUnit == 'L':
                    varUnit = 'mL'
                if varUnit == 'W':
                    varUnit = 'mW'
                if varUnit == 'Ci':
                    varUnit = 'mCi'
                if varUnit == 'gram/min' or varUnit == 'g/min':
                    varUnit = 'mg/min'
                if varUnit == 'gram/s' or varUnit == 'g/s':
                    varUnit = 'mg/s'
                if varUnit == 'gram/m3' or varUnit == 'g/m3':
                    varUnit = 'mg/m3'
                if varUnit == 'gram/L' or varUnit == 'g/L':
                    varUnit = 'mg/L'
                if varUnit == 'W/L':
                    varUnit = 'mW/L'
                if varUnit == 'Ci/L':
                    varUnit = 'mCi/L'
                if varUnit == 'L/min':
                    varUnit = 'mL/min'
                if varUnit == 'Pa':
                    varUnit = 'mPa'
                if varUnit == 's':
                    varUnit = 'ms'

            ax.set_xlabel('Time [' + timeUnit + ']', fontsize=9)
            ax.set_ylabel(varName + ' [' + varUnit + ']', fontsize=9)

            ymax = y.max()
            dy = ymax * .1
            ymax += dy
            ymin = y.min()
            ymin -= dy

            if abs(ymin - ymax) <= 1.e-4:
                ymin = -1.0
                ymax = 1.0

            ax.set_ylim(ymin, ymax)

            if nCols >= 4:
                for l in ax.get_xticklabels():
                    l.set_fontsize(8)
            else:
                for l in ax.get_xticklabels():
                    l.set_fontsize(10)
            for l in ax.get_yticklabels():
                l.set_fontsize(10)

            if timeUnit == 'h' and x.max() - x.min() <= 5.0:
                majorLocator = MultipleLocator(1.0)
                minorLocator = MultipleLocator(0.5)

                ax.xaxis.set_major_locator(majorLocator)
                ax.xaxis.set_minor_locator(minorLocator)

            if varScale == 'log' or varScale == 'log-log':
                ax.set_xscale('log')
                ax.set_yscale('log')
                positiveX = x > 0.0
                x = npy.extract(positiveX, x)
                y = npy.extract(positiveX, y)
                positiveY = y > 0.0
                x = npy.extract(positiveY, x)
                y = npy.extract(positiveY, y)
                if y.size > 0:
                    if y.min() > 0.0 and y.max() > y.min():
                        ymax = y.max()
                        dy = ymax * .1
                        ymax += dy
                        ymin = y.min()
                        ymin -= dy
                        if ymin < 0.0 or ymin > ymax / 1000.0:
                            ymin = ymax / 1000.0
                        ax.set_ylim(ymin, ymax)
                    else:
                        ax.set_ylim(1.0, 10.0)
                else:
                    ax.set_ylim(1.0, 10.0)

            if varScale == 'log-linear':
                ax.set_xscale('log')
                positiveX = x > 0.0
                x = npy.extract(positiveX, x)
                y = npy.extract(positiveX, y)

            if varScale == 'linear-log':
                ax.set_yscale('log')
                positiveY = y > 0.0
                x = npy.extract(positiveY, x)
                y = npy.extract(positiveY, y)
#      assert x.size == y.size, 'size error; stop.'
                if y.size > 0:
                    if y.min() > 0.0 and y.max() > y.min():
                        ymax = y.max()
                        dy = ymax * .1
                        ymax += dy
                        ymin = y.min()
                        ymin -= dy
                        if ymin < 0.0 or ymin > ymax / 1000.0:
                            ymin = ymax / 1000.0
                        ax.set_ylim(y.min(), ymax)
                    else:
                        ax.set_ylim(1.0, 10.0)
                else:
                    ax.set_ylim(1.0, 10.0)

        # ...................
        # make the plot here

            ax.plot(x, y, 's-', color='black', linewidth=0.5, markersize=2,
                    markeredgecolor='black', label=varLegend)

        # ...................

            ax.legend(loc='best', prop={'size': 7})

            s = '__plot_time_seq_dashboard(): plotted ' + varName + ' from ' + varLegend
            self.__log.debug(s)

    # end of: for iVar in range(nVar):

        figName = 'pyplot_' + str(self.__slot_id) + \
            '-timeseq-dashboard-' + str(iDash).zfill(2)
        fig.savefig(figName+'.png', dpi=200, fomat='png')
        plt.close(figNum)

        pickle.dump( fig, open(figName+'.pickle','wb') )

        s = '__plot_time_seq_dashboard(): created plot: ' + figName
        self.__log.debug(s)

        s = '__plot_time_seq_dashboard(): done with plotting'
        self.__log.debug(s)

    # clear __time_sequences_tmp
        self.__time_sequences_tmp = list()

        return

    def __plot_time_tables( self, initialTime=0.0, finalTime=0.0 ):

        nTimeSteps = len(self.__time_tables_data.keys())

        if nTimeSteps == 0:
            return

        s = '__plot_time_tables(): plotting tables'
        self.log.debug(s)

#  s = '__plot_time_tables(): keys = '+str(self.__time_tables_data.items())
#  self.log.debug(s)


#  assert nVariables <= 9, 'exceeded # of variables'

        fig = plt.figure('time-tables')

        gs = gridspec.GridSpec(2, 2)
        gs.update(left=0.1, right=0.98, wspace=0.4, hspace=0.4)

        axlst = list()
        axlst.append(fig.add_subplot(gs[0, 0]))
        axlst.append(fig.add_subplot(gs[0, 1]))
        axlst.append(fig.add_subplot(gs[1, 0]))
        axlst.append(fig.add_subplot(gs[1, 1]))
        axes = np.array(axlst)

        text = 'Cortix.PyPlot: Time-Tables Dashboard'
        fig.text(.5, .95, text, horizontalalignment='center', fontsize=16)

        for (key, val) in self.__time_tables_data.items():
            (timeStamp, timeUnit) = key
            columns = val

            if timeStamp == 0.0:
                continue

            distance = columns[0]
            xLabel = distance.get('name').strip()
            xUnit = distance.get('unit').strip()
            x = distance.text.strip().split(',')
            for i in range(len(x)):
                x[i] = float(x[i])
            x = np.array(x)

            for k in range(len(columns) - 1):

                ax = axes.flat[k]

                y = columns[k + 1]
                yLabel = y.get('name').strip()
                yUnit = y.get('unit').strip()
                yLegend = y.get('legend').strip()
                y = y.text.strip().split(',')
                for i in range(len(y)):
                    y[i] = float(y[i])
                y = np.array(y)

                if k == 0 or k == 1:
                    y *= 1000.0
                    yUnit = 'm'


#    x = data[:,0]
#    if x.max() >= 120.0:
#      x /= 60.0
#      if timeUnit == 'm': timeUnit = 'h'

#    y = data[:,1]
#    if y.max() >= 1000.0:
#      y /= 1000.0
#      if varUnit == 'gram' or varUnit == 'g':
#        varUnit = 'kg'
#    if y.max() <= .1:
#      y *= 1000.0
#      if varUnit == 'gram' or varUnit == 'g':
#        varUnit = 'mg'

                ax.set_xlabel(xLabel + ' [' + xUnit + ']', fontsize=9)
                ax.set_ylabel(yLabel + ' [' + yUnit + ']', fontsize=9)

#    ymax  = y.max()
#    dy    = ymax * .1
#    ymax += dy
#    ymin  = y.min()
#    ymin -= dy

#    ax.set_ylim(ymin,ymax)

                for l in ax.get_xticklabels():
                    l.set_fontsize(10)
                for l in ax.get_yticklabels():
                    l.set_fontsize(10)
#  color  = parameters['plot-color']
#  marker = parameters['plot-marker']+'-'
#  axis.set_xlim(2e-3, 2)
#  axis.set_aspect(1)
#  axis.set_title("adjustable = box")

                ax.plot(x, y, 's-', color='black', linewidth=0.5, markersize=2,
                        markeredgecolor='black', label=yLegend)

#      if k == 2 or k == 3: ax.set_yscale("log")


#      ax.legend( loc='best', prop={'size':8} )

    # end of for (varSpec,ax) in zip( self.timeSeriesData , axes.flat ):

        fig.savefig('pyplot-timetables.png', dpi=200, fomat='png')
        plt.close('time-tables')

        return

    def __read_manifest( self, xml_tree_file ):
        '''
        Parse the manifest
        '''

        assert isinstance(xml_tree_file, str)

        # Read the manifest 
        xml_tree = XMLTree( xml_tree_file=xml_tree_file )

        assert xml_tree.get_node_tag() == 'module_manifest'

        assert xml_tree.get_node_attribute('name') == 'pyplot'

        # List of (port_name, port_type, port_mode, port_multiplicity)
        __ports = list()

        self.__port_diagram = 'null-module-port-diagram'

        # Get manifest data  
        for child in xml_tree.get_node_children():
            (elem, tag, attributes, text) = child

            if tag == 'port':

                text = text.strip()

                assert len(attributes) == 3, "only <= 3 attributes allowed."

                tmp = dict()  # store port name and three attributes

                for attribute in attributes:
                    key = attribute[0].strip()
                    val = attribute[1].strip()

                    if key == 'type':
                        assert val == 'use' or val == 'provide' or val == 'input' or \
                            val == 'output', 'port attribute value invalid.'
                        tmp['port_name'] = text  # port_name
                        tmp['port_type'] = val   # port_type
                    elif key == 'mode':
                        file_value = val.split('.')[0]
                        assert file_value == 'file' or file_value == 'directory',\
                            'port attribute value invalid.'
                        tmp['port_mode'] = val
                    elif key == 'multiplicity':
                        tmp['port_multiplicity'] = int(val)  # port_multiplicity
                    else:
                        assert False, 'invalid port attribute: %r=%r. fatal.'%\
                                (key,val)

                assert len(tmp) == 4
                store = (tmp['port_name'], tmp['port_type'], tmp['port_mode'],
                         tmp['port_multiplicity'])

                # (port_name, port_type, port_mode, port_multiplicity)
                __ports.append(store)

                # clear
                tmp   = None
                store = None

            if tag == 'diagram':

                self.__port_diagram = text

            if tag == 'ascii_art':

                self.__ascii_art = text

        return

#======================= end class PyPlot: =======================================



# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://github.com/dpploy/...
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/COPYRIGHT
#
# Licensed under the GNU General Public License v. 3, please see LICENSE file.
# https://www.gnu.org/licenses/gpl-3.0.txt
#
from .src.cortix_main import Cortix
